from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.pointcloud_annotation.pointcloud_episode_tag import (
    PointcloudEpisodeTag,
)
from supervisely.pointcloud_annotation.pointcloud_tag_collection import (
    PointcloudTagCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudEpisodeTagCollection(PointcloudTagCollection):
    """
    Collection with :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag.PointcloudEpisodeTag` instances.
    :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag_collection.PointcloudEpisodeTagCollection` object is immutable.

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            # Create TagMetas (see class TagMeta for additional information about creating TagMetas)
            colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
            meta_car_color = sly.TagMeta('car color', sly.TagValueType.ONEOF_STRING, possible_values=colors)

            # Create tags
            tag_car_color_white = sly.PointcloudEpisodeTag(meta_car_color, value="white", frame_range=(15, 20))
            tag_car_color_red = sly.PointcloudEpisodeTag(meta_car_color, value="red", frame_range=(11, 22))
            tags_arr = [tag_car_color_white, tag_car_color_red]

            # Create PointcloudEpisodeTagCollection from point cloud episodes Tags
            tags = sly.PointcloudEpisodeTagCollection(tags_arr)

            # Add item to PointcloudEpisodeTagCollection
            meta_bus = sly.TagMeta('bus', sly.TagValueType.NONE)
            tag_bus = sly.PointcloudEpisodeTag(meta_bus, frame_range=(11, 14))

            # Remember that PointcloudEpisodeTagCollection is immutable, and we need to assign new instance of PointcloudEpisodeTagCollection to a new variable
            tags = tags.add(tag_bus)

            # You can also add multiple items to collection
            meta_truck = sly.TagMeta('truck', sly.TagValueType.NONE)
            meta_moto = sly.TagMeta('moto', sly.TagValueType.NONE)

            tag_truck = sly.PointcloudEpisodeTag(meta_truck, frame_range=(6, 10))
            tag_moto = sly.PointcloudEpisodeTag(meta_moto, frame_range=(11, 15))

            additional_tags = [tag_truck, tag_moto]

            tags = tags.add_items(additional_tags)

            # Has key, checks if given key exist in collection
            tags.has_key(meta_moto.key())
            # Output: True

            # Find intersection of given list of instances with collection items
            meta_bus = sly.TagMeta('bus', sly.TagValueType.NONE)
            tag_bus = sly.PointcloudEpisodeTag(meta_bus, frame_range=(0, 5))

            meta_truck = sly.TagMeta('truck', sly.TagValueType.NONE, frame_range=(6, 11))
            tag_truck = sly.PointcloudEpisodeTag(meta_truck)

            tags_vehicles = sly.PointcloudEpisodeTagCollection([tag_bus, tag_truck])

            tags_intersections = tags.intersection(tags_vehicles)
            print(tags_intersections.to_json())
            # Output: []

            # Let's add the moto Tag from another collection and compare them again
            tags_vehicles = tags_vehicles.add(tag_moto)

            tags_intersections = tags.intersection(tags_vehicles)
            print(tags_intersections.to_json())
            # Output: [
            #     {
            #         "name": "moto",
            #         "frameRange": [11, 15],
            #         "key": "3f37e10a658e440db3ab9f5c0be7fb67"
            #     }
            # ]

            # Find difference between collection and given list of Tags or PointcloudEpisodeTagCollection
            meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
            tag_car = sly.PointcloudEpisodeTag(meta_car, frame_range=(6, 11))

            meta_bicycle = sly.TagMeta('bicycle', sly.TagValueType.NONE)
            tag_bicycle = sly.PointcloudEpisodeTag(meta_bicycle, frame_range=(6, 11))

            tags_vehicles = sly.PointcloudEpisodeTagCollection([tag_car, tag_bicycle])

            meta_pedestrian = sly.TagMeta('pedestrian', sly.TagValueType.NONE)
            tag_pedestrian = sly.PointcloudEpisodeTag(meta_pedestrian, frame_range=(14, 20))

            meta_road = sly.TagMeta('road', sly.TagValueType.NONE)
            tag_road = sly.PointcloudEpisodeTag(meta_road, frame_range=(13, 22))

            difference = tags_vehicles.difference([tag_pedestrian, tag_road])
            print(difference.to_json())
            # Output: [
            #     {
            #         "name": "car",
            #         "frameRange": [6, 11],
            #         "key": "eba949cb996e44ac8aebec72a0875b13"
            #     },
            #     {
            #         "name": "bicycle",
            #         "frameRange": [6, 11],
            #         "key": "f9f03baec21e428ba04525d8df003e22"
            #     }
            # ]

            # Merge collection and given list of collections
            tags_vehicles = sly.PointcloudEpisodeTagCollection([tag_car, tag_bicycle])
            tags_merge = sly.PointcloudEpisodeTagCollection([tag_pedestrian, tag_road])

            merged_collections = tags_vehicles.merge(tags_merge)
            print(merged_collections.to_json())
            # Output: [
            #     {
            #         "frameRange": [6, 11],
            #         "key": "a6527453a3d34e68865bc0ad4b66f673",
            #         "name": "car"
            #     },
            #     {
            #         "frameRange": [6, 11],
            #         "key": "e9a9a306fd9c4ea180d1edf2a6fb1cf0",
            #         "name": "bicycle"
            #     },
            #     {
            #         "frameRange": [14, 20],
            #         "key": "8480532ff4db4600839ab403b1d0ab85",
            #         "name": "pedestrian"
            #     },
            #     {
            #         "frameRange": [13, 22],
            #         "key": "828f58c0c5174dbba40f7374dcc8016a",
            #         "name": "road"
            #     }
            # ]
    """

    item_type = PointcloudEpisodeTag

    def __iter__(self) -> Iterator[PointcloudEpisodeTag]:
        return next(self)

    @classmethod
    def from_api_response(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        id_to_tagmeta: Optional[Dict[int, TagMeta]] = None,
    ) -> PointcloudEpisodeTagCollection:
        """
        Create a PointcloudEpisodeTagCollection object from API response data.

        :param data: API response data.
        :type data: List[Dict]
        :param tag_meta_collection: Tag metadata collection.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param id_to_tagmeta: Mapping of tag IDs to tag metadata.
        :type id_to_tagmeta: Optional[Dict[int, :class:`~supervisely.annotation.tag_meta.TagMeta`]]
        :returns: Pointcloud episode tags collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag_collection.PointcloudEpisodeTagCollection`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 18428
                pcd_id = 19481209
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                pcd_info = api.pointcloud_episode.get_info_by_id(pcd_id)

                tag_collection = sly.PointcloudEpisodeTagCollection.from_api_response(
                    pcd_info.tags, project_meta.tag_metas
                )
                print(tag_collection)

                # Output:
                # Tags:
                # +-------+--------------+-------+
                # |  Name |  Value type  | Value |
                # +-------+--------------+-------+
                # | color | oneof_string |  red  |
                # |  car  |     none     |  None |
                # +-------+--------------+-------+
        """

        return super().from_api_response(data, tag_meta_collection, id_to_tagmeta=id_to_tagmeta)

    # def get_by_name(self, tag_name: str) -> List[PointcloudEpisodeTag]:
    #     return super().get_by_name(tag_name)

    # def get_single_by_name(self, tag_name: str) -> PointcloudEpisodeTag:
    #     return super().get_single_by_name(tag_name)

    @classmethod
    def from_json(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudEpisodeTagCollection:
        """
        Convert a list with dicts in json format to :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag_collection.PointcloudEpisodeTagCollection`.
        Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: :class:`List[Dict]`
        :param tag_meta_collection: Tag metadata collection.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud episode tags collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag_collection.PointcloudEpisodeTagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Initialize TagMetaCollection
                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                metas_arr = [meta_weather, meta_season]
                tag_metas = sly.TagMetaCollection(metas_arr)

                data = [
                    {"name":"Weather", "value":"Sunny"},
                    {"name":"Season", "value":"Spring"}
                ]

                tags = sly.PointcloudEpisodeTagCollection.from_json(data, tag_metas)
        """

        return super().from_json(data, tag_meta_collection, key_id_map=key_id_map)



# coding: utf-8
import numpy as np


class MultiClassAccuracy:
    """Compute multi-class accuracy for image segmentation."""

    def __init__(self, ignore_index: int=None, squeeze_targets: bool=True):
        """:param ignore_index: Target value that is ignored and does not contribute to accuracy.
        :type ignore_index: int
        :param squeeze_targets: Whether to squeeze targets array along 1st dimension.
        :type squeeze_targets: bool
        """
        self._ignore_index = ignore_index
        self._squeeze_targets = squeeze_targets

    def __call__(self, outputs, targets):
        outputs = outputs.data.cpu().numpy()
        outputs = np.argmax(outputs, axis=1)

        targets = targets.data.cpu().numpy()
        if self._squeeze_targets:
            targets = np.squeeze(targets, 1)  # 4d to 3d

        total_pixels = np.prod(outputs.shape)
        if self._ignore_index is not None:
            total_pixels -= np.sum((targets == self._ignore_index).astype(int))
        correct_pixels = np.sum((outputs == targets).astype(int))
        res = correct_pixels / total_pixels
        return res



# coding: utf-8


# docs

from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from shapely.geometry import LineString
from shapely.geometry import Polygon as ShapelyPolygon
from shapely.geometry import mapping

from supervisely.sly_logger import logger
from supervisely.geometry import validation
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    EXTERIOR,
    ID,
    LABELER_LOGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.geometry.conversions import shapely_figure_to_coords_list
from supervisely.geometry.point import PointLocation
from supervisely.geometry.point_location import row_col_list_to_points
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.vector_geometry import VectorGeometry


class Polyline(VectorGeometry):
    """Open 2D polyline (sequence of connected line segments). Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return "line"

    def __init__(
        self,
        exterior: Union[List[PointLocation], List[List[int, int]], List[Tuple[int, int]]],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        Polyline geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.polyline.Polyline` object is immutable.

        :param exterior: List of exterior coordinates, the Polyline is defined with these points.
        :type exterior: List[:class:`~supervisely.geometry.point_location.PointLocation`], List[List[int, int]], List[Tuple[int, int]
        :param sly_id: Polyline ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Polyline belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Polyline.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Polyline was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Polyline was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :raises ValueError: field exterior must contain at least two points to create Polyline.

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                exterior = [sly.PointLocation(730, 2104), sly.PointLocation(2479, 402), sly.PointLocation(1500, 780)]
                # or exterior = [[730, 2104], [2479, 402], [1500, 780]]
                # or exterior = [(730, 2104), (2479, 402), (1500, 780)]
                figure = sly.Polyline(exterior)
        """
        if len(exterior) < 2:
            raise ValueError(
                f'"{EXTERIOR}" field must contain at least two points to create "Polyline" object.'
            )

        super().__init__(
            exterior=exterior,
            interior=[],
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    @classmethod
    def from_json(cls, data: Dict) -> Polyline:
        """
        Convert a json dict to Polyline. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Polyline in json format as a dict.
        :type data: dict
        :returns: Polyline from json.
        :rtype: :class:`~supervisely.geometry.polyline.Polyline`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                figure_json = {
                    "points": {
                        "exterior": [
                            [2104, 730],
                            [402, 2479],
                            [780, 1500]
                        ],
                        "interior": []
                    }
                }
                figure = sly.Polyline.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            exterior=row_col_list_to_points(data[POINTS][EXTERIOR], flip_row_col_order=True),
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def crop(self, rect: Rectangle) -> List[Polyline]:
        """
        Crops current Polyline.

        :param rect: Rectangle to crop Polyline from.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of Polygons from Rectangle.
        :rtype: List[:class:`~supervisely.geometry.polyline.Polyline`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                crop_figures = figure.crop(sly.Rectangle(0, 0, 100, 200))
        """
        try:
            clipping_window = [
                [rect.top, rect.left],
                [rect.top, rect.right],
                [rect.bottom, rect.right],
                [rect.bottom, rect.left],
            ]
            clipping_window_shpl = ShapelyPolygon(clipping_window)

            exterior = self.exterior_np
            intersections_polygon = LineString(exterior).intersection(clipping_window_shpl)
            mapping_shpl = mapping(intersections_polygon)
        except Exception:
            logger.warn("Line cropping exception, shapely.", exc_info=False)
            raise

        res_lines_pts = shapely_figure_to_coords_list(mapping_shpl)

        # tiny hack to combine consecutive segments
        lines_combined = []
        if res_lines_pts != [()]:
            for simple_l in res_lines_pts:
                if len(lines_combined) > 0:
                    prev = lines_combined[-1]
                    if prev[-1] == simple_l[0]:
                        lines_combined[-1] = list(prev) + list(simple_l[1:])
                        continue
                lines_combined.append(simple_l)

        return [Polyline(row_col_list_to_points(line)) for line in lines_combined]

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """ """
        self._draw_contour_impl(bitmap, color, thickness, config=config)

    def _draw_contour_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """ """
        exterior = self.exterior_np[:, ::-1]
        cv2.polylines(bitmap, pts=[exterior], isClosed=False, color=color, thickness=thickness)

    @property
    def area(self) -> float:
        """
        Polyline area, always 0.0.

        :returns: Area of current Polyline, always 0.0
        :rtype: float

        :Usage Example:

            .. code-block:: python

                print(figure.area)
                # Output: 0.0
        """
        return 0.0

    def approx_dp(self, epsilon: float) -> Polyline:
        """
        Approximates a Polyline with the specified precision.

        :param epsilon: Specifying the approximation accuracy.
        :type epsilon: float
        :returns: Approximated Polyline.
        :rtype: :class:`~supervisely.geometry.polyline.Polyline`

        :Usage Example:

            .. code-block:: python

                # Remember that Polyline class object is immutable, and we need to assign new instance of Polyline to a new variable
                approx_figure = figure.approx_dp(0.75)
        """
        exterior_np = self._approx_ring_dp(self.exterior_np, epsilon, closed=True).tolist()
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        return Polyline(exterior)

    @classmethod
    def allowed_transforms(cls):
        """
        Returns the allowed transforms for the Polyline.
        """
        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon
        from supervisely.geometry.rectangle import Rectangle

        return [AnyGeometry, Rectangle, Bitmap, Polygon, AlphaMask]



from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from supervisely import logger
from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.image_api import ImageInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.app import DataJson
from supervisely.app.widgets import Button, Card, Progress, Stepper, Text, Widget
from supervisely.nn.model.prediction import Prediction
from supervisely.project.project import ProjectType
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.video_project import VideoInfo
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.video_object import VideoObject
from supervisely.video_annotation.video_object_collection import VideoObjectCollection

button_clicked = {}


def update_custom_params(
    button: Button,
    params_dct: Dict[str, Any],
) -> None:
    button_state = button.get_json_data()
    for key in params_dct.keys():
        if key not in button_state:
            raise AttributeError(f"Parameter {key} doesn't exists.")
        else:
            DataJson()[button.widget_id][key] = params_dct[key]
    DataJson().send_changes()


def update_custom_button_params(
    button: Button,
    params_dct: Dict[str, Any],
) -> None:
    params = params_dct.copy()
    if "icon" in params and params["icon"] is not None:
        new_icon = f'<i class="{params["icon"]}" style="margin-right: {button._icon_gap}px"></i>'
        params["icon"] = new_icon
    update_custom_params(button, params)


def disable_enable(widgets: List[Widget], disable: bool = True):
    for w in widgets:
        if disable:
            w.disable()
        else:
            w.enable()


def unlock_lock(cards: List[Card], unlock: bool = True, message: str = None):
    for w in cards:
        if unlock:
            w.unlock()
            # w.uncollapse()
        else:
            w.lock(message)
            # w.collapse()


def collapse_uncollapse(cards: List[Card], collapse: bool = True):
    for w in cards:
        if collapse:
            w.collapse()
        else:
            w.uncollapse()


def wrap_button_click(
    button: Button,
    cards_to_unlock: List[Card],
    widgets_to_disable: List[Widget],
    callback: Optional[Callable] = None,
    lock_msg: str = None,
    upd_params: bool = True,
    validation_text: Text = None,
    validation_func: Optional[Callable] = None,
    on_select_click: Optional[Callable] = None,
    on_reselect_click: Optional[Callable] = None,
    collapse_card: Tuple[Card, bool] = None,
) -> Callable[[Optional[bool]], None]:
    global button_clicked

    select_params = {"icon": None, "plain": False, "text": "Select"}
    reselect_params = {"icon": "zmdi zmdi-refresh", "plain": True, "text": "Reselect"}
    bid = button.widget_id
    button_clicked[bid] = False

    def button_click(button_clicked_value: Optional[bool] = None, suppress_actions: bool = False):
        if button_clicked_value is None or button_clicked_value is False:
            if validation_func is not None:
                success = validation_func()
                if not success:
                    return

        if button_clicked_value is not None:
            button_clicked[bid] = button_clicked_value
        else:
            button_clicked[bid] = not button_clicked[bid]

        if button_clicked[bid] and upd_params:
            update_custom_button_params(button, reselect_params)
            if not suppress_actions and on_select_click is not None:
                for func in on_select_click:
                    func()
        else:
            update_custom_button_params(button, select_params)
            if not suppress_actions and on_reselect_click is not None:
                for func in on_reselect_click:
                    func()
            validation_text.hide()

        unlock_lock(
            cards_to_unlock,
            unlock=button_clicked[bid],
            message=lock_msg,
        )
        disable_enable(
            widgets_to_disable,
            disable=button_clicked[bid],
        )
        if callback is not None and not button_clicked[bid]:
            callback(False, True)

        if collapse_card is not None:
            card, collapse = collapse_card
            if collapse:
                collapse_uncollapse([card], collapse)

    return button_click


def set_stepper_step(stepper: Stepper, button: Button, next_pos: int):
    bid = button.widget_id
    if button_clicked[bid] is True:
        stepper.set_active_step(next_pos)
    else:
        stepper.set_active_step(next_pos - 1)


def find_parents_in_tree(
    tree: Dict[DatasetInfo, Dict], dataset_id: int, with_self: bool = False
) -> Optional[List[DatasetInfo]]:
    """
    Find all parent datasets in the tree for a given dataset ID.
    """

    def _dfs(subtree: Dict[DatasetInfo, Dict], parents: List[DatasetInfo]):
        for dataset_info, children in subtree.items():
            if dataset_info.id == dataset_id:
                if with_self:
                    return parents + [dataset_info]
                return parents
            res = _dfs(children, parents + [dataset_info])
            if res is not None:
                return res
        return None

    return _dfs(tree, [])


def _copy_items_to_dataset(
    api: Api,
    src_dataset_id: int,
    dst_dataset: DatasetInfo,
    project_type: str,
    with_annotations: bool = True,
    progress_cb: Callable = None,
    progress: Progress = None,
    items_infos: List[Union[ImageInfo, VideoInfo]] = None,
) -> Union[List[ImageInfo], List[VideoInfo]]:
    if progress is None:
        progress = Progress()

    def combined_progress(n):
        progress_cb(n)
        pbar.update(n)

    if project_type == ProjectType.IMAGES:
        if items_infos is None:
            items_infos = api.image.get_list(src_dataset_id)
        with progress(
            message=f"Copying items from dataset: {dst_dataset.name}", total=len(items_infos)
        ) as pbar:

            if progress_cb:
                _progress_cb = combined_progress
            else:
                _progress_cb = pbar.update

            progress.show()
            copied = api.image.copy_batch_optimized(
                src_dataset_id=src_dataset_id,
                src_image_infos=items_infos,
                dst_dataset_id=dst_dataset.id,
                with_annotations=with_annotations,
                progress_cb=_progress_cb,
            )
            progress.hide()
    elif project_type == ProjectType.VIDEOS:
        if items_infos is None:
            items_infos = api.video.get_list(src_dataset_id)

        with progress(
            message=f"Copying items from dataset: {dst_dataset.name}", total=len(items_infos)
        ) as pbar:
            if progress_cb:
                _progress_cb = combined_progress
            else:
                _progress_cb = pbar.update
            progress.show()
            copied = api.video.copy_batch(
                dst_dataset_id=dst_dataset.id,
                ids=[info.id for info in items_infos],
                with_annotations=with_annotations,
                progress_cb=_progress_cb,
            )
            progress.hide()
    else:
        raise NotImplementedError(f"Copy not implemented for project type {project_type}")
    return copied


def get_items_infos(
    api: Api, items_ids: List[int], project_type: str
) -> List[Union[ImageInfo, VideoInfo]]:
    if project_type == ProjectType.IMAGES:
        items_infos: List[ImageInfo] = api.image.get_info_by_id_batch(items_ids)
    elif project_type == ProjectType.VIDEOS:
        items_infos: List[VideoInfo] = api.video.get_info_by_id_batch(items_ids)
    else:
        raise NotImplementedError(f"Items of type {project_type} are not supported")
    return items_infos


def copy_items_to_project(
    api: Api,
    src_project_id: int,
    items: Union[List[ImageInfo], List[VideoInfo]],
    dst_project_id: int,
    with_annotations: bool = True,
    progress_cb: Progress = None,
    ds_progress: Progress = None,
    project_type: str = None,
    src_datasets_tree: Dict[DatasetInfo, Dict] = None,
) -> Union[List[ImageInfo], List[VideoInfo]]:
    if project_type is None:
        dst_project_info = api.project.get_info_by_id(src_project_id)
        project_type = dst_project_info.type
    if len(items) == 0:
        return []
    if len(set(info.project_id for info in items)) != 1:
        raise ValueError("Items must belong to the same project")

    items_by_dataset: Dict[int, List[Union[ImageInfo, VideoInfo]]] = {}
    for item_info in items:
        items_by_dataset.setdefault(item_info.dataset_id, []).append(item_info)

    if src_datasets_tree is None:
        src_datasets_tree = api.dataset.get_tree(src_project_id)

    created_datasets: Dict[int, DatasetInfo] = {}
    processed_copy: Set[int] = set()

    copied_items = {}
    for dataset_id, items_infos in items_by_dataset.items():
        chain = find_parents_in_tree(src_datasets_tree, dataset_id, with_self=True)
        if not chain:
            logger.warning(f"Dataset id {dataset_id} not found in project. Skipping")
            continue

        parent_created_id = None
        for ds_info in chain:
            if ds_info.id in created_datasets:
                parent_created_id = created_datasets[ds_info.id].id
                continue

            created_ds = api.dataset.create(
                dst_project_id,
                ds_info.name,
                description=ds_info.description,
                change_name_if_conflict=False,
                parent_id=parent_created_id,
            )
            if ds_info.custom_data:
                created_ds = api.dataset.update_custom_data(created_ds.id, ds_info.custom_data)
            created_datasets[ds_info.id] = created_ds
            parent_created_id = created_ds.id

        if dataset_id not in processed_copy:
            copied_ds_items = _copy_items_to_dataset(
                api=api,
                src_dataset_id=dataset_id,
                dst_dataset=created_datasets[dataset_id],
                project_type=project_type,
                with_annotations=with_annotations,
                progress_cb=progress_cb,
                progress=ds_progress,
                items_infos=items_infos,
            )
            for src_info, dst_info in zip(items_infos, copied_ds_items):
                copied_items[src_info.id] = dst_info
            processed_copy.add(dataset_id)
    return [copied_items[item.id] for item in items]


def create_project(
    api: Api,
    project_id: int,
    project_name: str,
    workspace_id: int,
    copy_meta: bool = False,
    project_type: str = None,
) -> ProjectInfo:
    if project_type is None:
        project_info = api.project.get_info_by_id(project_id)
        project_type = project_info.type
    created_project = api.project.create(
        workspace_id,
        project_name,
        type=project_type,
        change_name_if_conflict=True,
    )
    if copy_meta:
        api.project.merge_metas(src_project_id=project_id, dst_project_id=created_project.id)
    return created_project


def copy_project(
    api: Api,
    project_id: int,
    workspace_id: int,
    project_name: str,
    items_ids: List[int] = None,
    with_annotations: bool = True,
    progress: Progress = None,
) -> ProjectInfo:
    dst_project = create_project(
        api, project_id, project_name, workspace_id=workspace_id, copy_meta=True
    )
    items = []
    if items_ids is None:
        project_type = dst_project.type
        datasets = api.dataset.get_list(project_id, recursive=True)
        if project_type == ProjectType.IMAGES:
            get_items_f = api.image.get_list
        elif project_type == ProjectType.VIDEOS:
            get_items_f = api.video.get_list
        else:
            raise NotImplementedError(f"Project type {project_type} is not supported")
        for ds in datasets:
            ds_items = get_items_f(dataset_id=ds.id)
            if ds_items:
                items.extend(ds_items)
    else:
        items = get_items_infos(api, items_ids, dst_project.type)
    copy_items_to_project(
        api=api,
        src_project_id=project_id,
        items=items,
        dst_project_id=dst_project.id,
        with_annotations=with_annotations,
        ds_progress=progress,
        project_type=dst_project.type,
    )
    return dst_project


def video_annotation_from_predictions(
    predictions: List[Prediction], project_meta: ProjectMeta, frame_size: Tuple[int, int]
) -> VideoAnnotation:
    objects = {}
    frames = []
    for i, prediction in enumerate(predictions):
        figures = []
        for label in prediction.annotation.labels:
            obj_name = label.obj_class.name
            if not obj_name in objects:
                obj_class = project_meta.get_obj_class(obj_name)
                if obj_class is None:
                    continue
                objects[obj_name] = VideoObject(obj_class)

            vid_object = objects[obj_name]
            if vid_object:
                figures.append(VideoFigure(vid_object, label.geometry, frame_index=i))
        frame = Frame(i, figures=figures)
        frames.append(frame)
    return VideoAnnotation(
        img_size=frame_size,
        frames_count=len(frames),
        objects=VideoObjectCollection(list(objects.values())),
        frames=FrameCollection(frames),
    )



from typing import List, Union
from supervisely import Api, DatasetInfo
from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget


class MatchDatasets(Widget):
    """Widget for matching items between two dataset lists; computes statistics and reports matches."""

    def __init__(
        self,
        left_datasets: List[DatasetInfo] = None,
        right_datasets: List[DatasetInfo] = None,
        left_name=None,
        right_name=None,
        widget_id=None,
    ):
        """
        :param left_datasets: Left dataset list.
        :type left_datasets: List[DatasetInfo], optional
        :param right_datasets: Right dataset list.
        :type right_datasets: List[DatasetInfo], optional
        :param left_name: Label for left column.
        :param right_name: Label for right column.
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._left_ds = left_datasets
        self._right_ds = right_datasets
        self._left_name = "Left Datasets" if left_name is None else left_name
        self._right_name = "Right Datasets" if right_name is None else right_name
        self._api = Api()

        self._done = False

        self._results = None
        self._stat = {}

        super().__init__(widget_id=widget_id, file_path=__file__)

        if self._left_ds is not None and self._right_ds is not None:
            self._load_datasets_statistic()

    def get_json_data(self):
        return {"left_name": self._left_name, "right_name": self._right_name}

    def get_json_state(self):
        return {"done": self._done}

    def set(
        self,
        left_datasets: List[DatasetInfo] = None,
        right_datasets: List[DatasetInfo] = None,
        left_name=None,
        right_name=None,
    ):
        self._left_ds = left_datasets
        self._right_ds = right_datasets
        self._left_name = self._left_name if left_name is None else left_name
        self._right_name = self._right_name if right_name is None else right_name

        self._done = False
        self._loading = False

        self._results = None
        self._stat = {}

        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()

        if self._left_ds is not None and self._right_ds is not None:
            self._load_datasets_statistic()

    def get_stat(self):
        return self._stat

    def _load_datasets_statistic(self):
        self._stat = {}
        ds_info1, ds_images1 = self._get_all_images(self._left_ds)
        ds_info2, ds_images2 = self._get_all_images(self._right_ds)
        result = self._process_items(ds_info1, ds_images1, ds_info2, ds_images2)

        DataJson()[self.widget_id]["table"] = result
        DataJson().send_changes()
        StateJson()[self.widget_id]["done"] = True
        StateJson().send_changes()

    def _process_items(self, ds_info1, collection1, ds_info2, collection2):
        ds_names = ds_info1.keys() | ds_info2.keys()

        results = []
        for idx, name in enumerate(ds_names):
            compare = {"dsIndex": idx}
            images1 = collection1.get(name, [])
            images2 = collection2.get(name, [])
            if len(images1) == 0:
                compare["message"] = ["unmatched (in GT project)"]
                compare["icon"] = [
                    ["zmdi zmdi-long-arrow-left", "zmdi zmdi-alert-circle-o"]
                ]

                compare["color"] = ["#F39C12"]
                compare["numbers"] = [-1]
                compare["left"] = {"name": ""}
                compare["right"] = {"name": name, "count": len(images2)}
                self._stat[name] = {"dataset_matched": "right"}
            elif len(images2) == 0:
                compare["message"] = ["unmatched (in PRED project)"]
                compare["icon"] = [
                    ["zmdi zmdi-alert-circle-o", "zmdi zmdi-long-arrow-right"]
                ]
                compare["color"] = ["#F39C12"]
                compare["numbers"] = [-1]
                compare["left"] = {"name": name, "count": len(images1)}
                compare["right"] = {"name": ""}
                self._stat[name] = {"dataset_matched": "left"}
            else:
                img_dict1 = {img_info.name: img_info for img_info in images1}
                img_dict2 = {img_info.name: img_info for img_info in images2}
                matched = []
                diff = []
                same_names = img_dict1.keys() & img_dict2.keys()
                for img_name in same_names:
                    dest = (
                        matched
                        if img_dict1[img_name].hash == img_dict2[img_name].hash
                        else diff
                    )
                    dest.append(
                        {"left": img_dict1[img_name], "right": img_dict2[img_name]}
                    )

                uniq1 = [img_dict1[name] for name in img_dict1.keys() - same_names]
                uniq2 = [img_dict2[name] for name in img_dict2.keys() - same_names]

                compare["message"] = [
                    "matched",
                    "conflicts",
                    "unique (left)",
                    "unique (right)",
                ]
                compare["icon"] = [
                    ["zmdi zmdi-check"],
                    ["zmdi zmdi-close"],
                    ["zmdi zmdi-plus-circle-o"],
                    ["zmdi zmdi-plus-circle-o"],
                ]
                compare["color"] = ["green", "red", "#20a0ff", "#20a0ff"]
                compare["numbers"] = [len(matched), len(diff), len(uniq1), len(uniq2)]
                compare["left"] = {"name": name, "count": len(images1)}
                compare["right"] = {"name": name, "count": len(images2)}

                self._stat[name] = {
                    "dataset_matched": "both",
                    "matched": matched,
                    "conflicts": diff,
                    "unique_left": uniq1,
                    "unique_right": uniq2,
                }
            results.append(compare)

        self._results = results
        return results

    def _get_all_images(self, datasets):
        ds_info = {}
        ds_images = {}
        for dataset in datasets:
            ds_info[dataset.name] = dataset
            images = self._api.image.get_list(dataset.id)
            ds_images[dataset.name] = images
        return (
            ds_info,
            ds_images,
        )



from typing import Dict, Literal, Optional, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Button, Widget


class Editor(Widget):
    """Code editor with syntax highlighting."""

    def __init__(
        self,
        initial_text: Optional[str] = "",
        height_px: Optional[int] = 100,
        height_lines: Optional[int] = None,
        language_mode: Optional[Literal["json", "html", "plain_text", "yaml", "python"]] = "json",
        readonly: Optional[bool] = False,
        show_line_numbers: Optional[bool] = True,
        highlight_active_line: Optional[bool] = True,
        restore_default_button: Optional[bool] = True,
        widget_id: Optional[int] = None,
        auto_format: bool = False,
    ):
        """
        :param initial_text: Initial content.
        :type initial_text: Optional[str]
        :param height_px: Height in pixels.
        :type height_px: Optional[int]
        :param height_lines: Height in lines (overrides height_px; >=1000 shows all).
        :type height_lines: Optional[int]
        :param language_mode: Language mode of the editor, one of: json, html, plain_text, yaml, python.
        :type language_mode: Optional[Literal["json", "html", "plain_text", "yaml", "python"]]
        :param readonly: Read-only mode.
        :type readonly: Optional[bool]
        :param show_line_numbers: Show line numbers.
        :type show_line_numbers: Optional[bool]
        :param highlight_active_line: Highlight current line.
        :type highlight_active_line: Optional[bool]
        :param restore_default_button: Show restore button.
        :type restore_default_button: Optional[bool]
        :param auto_format: Auto-format JSON on init.
        :type auto_format: bool, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Editor
                editor = Editor(initial_text="print('hi')", language_mode="python")
        """
        self._initial_code = initial_text
        self._current_code = initial_text
        self._height_px = height_px
        self._height_lines = height_lines
        self._language_mode = language_mode
        self._readonly = readonly
        self._show_line_numbers = show_line_numbers
        self._highlight_active_line = highlight_active_line
        self._restore_button = None
        self._auto_format = auto_format

        if restore_default_button:
            self._restore_button = Button("Restore Default", button_type="text", plain=True)

            @self._restore_button.click
            def restore_default():
                self._current_code = self._initial_code
                StateJson()[self.widget_id]["text"] = self._current_code
                StateJson().send_changes()

        super(Editor, self).__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[bool, str, int]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - editor_options: Dictionary with editor options
                - height: Height of the editor in pixels
                - mode: Language mode of the editor, one of: json, html, plain_text, yaml, python
                - readOnly: If True, editor will be readonly
                - showGutter: If True, line numbers will be shown
                - maxLines: Overwrites height if specified. If >= 1000, all lines will be displayed
                - highlightActiveLine: If True, active line will be highlighted

        :returns: Dictionary with widget data
        :rtype: Dict[str, Union[bool, str, int]]
        """
        return {
            "editor_options": {
                "height": f"{self._height_px}px",
                "mode": f"ace/mode/{self._language_mode}",
                "readOnly": self._readonly,
                "showGutter": self._show_line_numbers,
                "maxLines": self._height_lines,
                "highlightActiveLine": self._highlight_active_line,
                "formatJsonOnInit": self._auto_format,
            },
        }

    def get_json_state(self) -> Dict[str, str]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - text: current text in the editor

        :returns: Dictionary with widget state
        :rtype: Dict[str, str]
        """
        return {"text": self._current_code}

    def get_text(self) -> str:
        """Returns current text in the editor.
        Same as get_value().

        :returns: current text in the editor
        :rtype: str
        """
        return StateJson()[self.widget_id]["text"]

    def get_value(self) -> str:
        """Returns current text in the editor.
        Same as get_text().

        :returns: current text in the editor
        :rtype: str
        """
        return StateJson()[self.widget_id]["text"]

    def set_text(
        self,
        text: Optional[str] = "",
        language_mode: Optional[Literal["json", "html", "plain_text", "yaml", "python"]] = None,
    ) -> None:
        """Sets current text in the editor.
        This method overwrites the current text in the editor.

        :param text: current text in the editor
        :type text: Optional[str]
        :param language_mode: Language mode of the editor, one of: json, html, plain_text, yaml, python
        :type language_mode: Optional[Literal["json", "html", "plain_text", "yaml", "python"]]
        """

        self._initial_code = text
        self._current_code = text
        self._language_mode = language_mode
        StateJson()[self.widget_id]["text"] = text
        StateJson().send_changes()
        if language_mode is not None:
            self._language_mode = f"ace/mode/{language_mode}"
            DataJson()[self.widget_id]["editor_options"]["mode"] = self._language_mode
            DataJson().send_changes()

    @property
    def readonly(self) -> bool:
        """Returns True if editor is readonly, False otherwise.

        :returns: True if editor is readonly, False otherwise
        :rtype: bool
        """
        return self._readonly

    @readonly.setter
    def readonly(self, value: bool) -> None:
        """Sets editor readonly mode.

        :param value: True if editor is readonly, False otherwise
        :type value: bool
        """
        self._readonly = value
        DataJson()[self.widget_id]["editor_options"]["readOnly"] = self._readonly
        DataJson().send_changes()

    def show_line_numbers(self) -> None:
        """Enables line numbers in the editor."""
        self._show_line_numbers = True
        DataJson()[self.widget_id]["editor_options"]["showGutter"] = self._show_line_numbers
        DataJson().send_changes()

    def hide_line_numbers(self):
        """Disables line numbers in the editor."""
        self._show_line_numbers = False
        DataJson()[self.widget_id]["editor_options"]["showGutter"] = self._show_line_numbers
        DataJson().send_changes()



import datetime
import tempfile
import threading
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import pandas as pd

from supervisely._utils import logger
from supervisely.api.api import Api
from supervisely.api.app_api import ModuleInfo
from supervisely.app.widgets.agent_selector.agent_selector import AgentSelector
from supervisely.app.widgets.button.button import Button
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.ecosystem_model_selector.ecosystem_model_selector import (
    EcosystemModelSelector,
)
from supervisely.app.widgets.experiment_selector.experiment_selector import (
    ExperimentSelector,
)
from supervisely.app.widgets.fast_table.fast_table import FastTable
from supervisely.app.widgets.field.field import Field
from supervisely.app.widgets.flexbox.flexbox import Flexbox
from supervisely.app.widgets.model_info.model_info import ModelInfo
from supervisely.app.widgets.radio_tabs.radio_tabs import RadioTabs
from supervisely.app.widgets.text.text import Text
from supervisely.app.widgets.widget import Widget
from supervisely.io import env
from supervisely.nn.experiments import ExperimentInfo, get_experiment_infos
from supervisely.nn.model.model_api import ModelAPI


class DeployModel(Widget):
    """UI widget to deploy or connect to a model and return a ready-to-use :class:`~supervisely.nn.model.model_api.ModelAPI`."""

    class DeployMode:
        """Strategy interface for different deployment sources (connect, pretrained, custom)."""

        def deploy(self, agent_id: int = None) -> ModelAPI:
            raise NotImplementedError("This method should be implemented in subclasses.")

        def get_deploy_parameters(self) -> Dict[str, Any]:
            raise NotImplementedError("This method should be implemented in subclasses.")

        def load_from_json(self, data: Dict[str, Any]) -> None:
            raise NotImplementedError("This method should be implemented in subclasses.")

        @property
        def layout(self) -> Widget:
            raise NotImplementedError("This property should be implemented in subclasses.")

    class Connect(DeployMode):
        """Deploy mode that connects to an already deployed model session."""

        class COLUMN:
            """Column names for the deployed sessions table."""
            SESSION_ID = "Session ID"
            APP_NAME = "App Name"
            FRAMEWORK = "Framework"
            MODEL = "Model"

        COLUMNS = [
            str(COLUMN.SESSION_ID),
            str(COLUMN.APP_NAME),
            str(COLUMN.FRAMEWORK),
            str(COLUMN.MODEL),
        ]

        def __init__(self, deploy_model: "DeployModel"):
            """
            :param deploy_model: Parent DeployModel widget.
            :type deploy_model: DeployModel, optional
            """
            self.api = deploy_model.api
            self.team_id = deploy_model.team_id
            self._cache = deploy_model._cache
            self.deploy_model = deploy_model
            self._layout = self._create_layout()
            self._update_sessions()

        def _create_layout(self) -> Container:
            self.refresh_button = Button(
                "",
                icon="zmdi zmdi-refresh",
                button_type="text",
            )
            self.sessions_table = FastTable(
                columns=self.COLUMNS,
                page_size=10,
                is_radio=True,
                header_left_content=self.refresh_button,
            )

            @self.refresh_button.click
            def _refresh_button_clicked():
                self._update_sessions()

            return self.sessions_table

        @property
        def layout(self) -> FastTable:
            return self._layout

        def _data_from_session(self, session: Dict) -> Dict[str, Any]:
            task_info = session["task_info"]
            deploy_info = session["model_info"]
            return {
                self.COLUMN.SESSION_ID: task_info["id"],
                self.COLUMN.APP_NAME: task_info["meta"]["app"]["name"],
                self.COLUMN.FRAMEWORK: self.deploy_model._framework_from_task_info(task_info),
                self.COLUMN.MODEL: deploy_info["model_name"],
            }

        def _update_sessions(self) -> None:
            self.sessions_table.loading = True
            try:
                self.sessions_table.clear()
                sessions = self.api.nn.list_deployed_models(team_id=self.team_id)
                data = [self._data_from_session(session) for session in sessions]
                df = pd.DataFrame.from_records(data=data, columns=self.COLUMNS)
                self.sessions_table.read_pandas(df)
                if len(data) == 0:
                    self.deploy_model.connect_button.disable()
                else:
                    self.deploy_model.connect_button.enable()
            except Exception as e:
                logger.error(
                    f"Failed to load deployed models: {e}",
                    exc_info=True,
                )
            finally:
                self.sessions_table.loading = False

        def deploy(self, agent_id: int = None) -> ModelAPI:
            deploy_parameters = self.get_deploy_parameters()
            logger.info(f"Connecting to model with parameters:", extra=deploy_parameters)
            session_id = deploy_parameters["session_id"]
            model_api = self.api.nn.connect(task_id=session_id)
            return model_api

        def get_deploy_parameters(self) -> Dict[str, Any]:
            selected_row = self.sessions_table.get_selected_row()
            return {
                "session_id": selected_row.row[self.COLUMNS.index(str(self.COLUMN.SESSION_ID))],
            }

        def load_from_json(self, data: Dict):
            session_id = data["session_id"]
            self._update_sessions()
            self.sessions_table.select_row_by_value(str(self.COLUMN.SESSION_ID), session_id)

    class Pretrained(DeployMode):
        """Deploy mode that deploys a pretrained model from the ecosystem model catalog."""
        class COLUMN:
            # TODO: columns are the same as in EcosystemModelSelector, make a common base class
            """Column names for the pretrained model selector table."""
            FRAMEWORK = "Framework"
            MODEL_NAME = "Model"
            TASK_TYPE = "Task Type"
            PARAMETERS = "Parameters (M)"
            # TODO: support metrics for different tasks
            MAP = "mAP"

        COLUMNS = [
            str(COLUMN.FRAMEWORK),
            str(COLUMN.MODEL_NAME),
            str(COLUMN.TASK_TYPE),
            str(COLUMN.PARAMETERS),
            str(COLUMN.MAP),
        ]

        def __init__(self, deploy_model: "DeployModel"):
            """
            :param deploy_model: Parent DeployModel widget.
            :type deploy_model: DeployModel, optional
            """
            self.api = deploy_model.api
            self.team_id = deploy_model.team_id
            self._cache = deploy_model._cache
            self.deploy_model = deploy_model
            self._model_api = None
            self._last_selected_framework = None
            self._layout = self._create_layout()

        @property
        def layout(self) -> FastTable:
            return self._layout

        def _create_layout(self) -> Container:
            self.model_selector = EcosystemModelSelector(api=self.api)
            return self.model_selector

        def get_deploy_parameters(self) -> Dict[str, Any]:
            selected_model = self.model_selector.get_selected()
            return {
                "framework": selected_model["framework"],
                "model_name": selected_model["name"],
            }

        def load_from_json(self, data: Dict[str, Any]) -> None:
            framework = data["framework"]
            model_name = data["model_name"]
            self.model_selector.select_framework_and_model_name(framework, model_name)

        def deploy(self, agent_id: int = None) -> ModelAPI:
            deploy_parameters = self.get_deploy_parameters()
            logger.info(f"Deploying pretrained model with parameters:", extra=deploy_parameters)
            framework = deploy_parameters["framework"]
            model_name = deploy_parameters["model_name"]
            model_api = self.api.nn.deploy(model=f"{framework}/{model_name}", agent_id=agent_id)
            return model_api

    class Custom(DeployMode):
        """Deploy mode that deploys a model from a user experiment checkpoint."""
        def __init__(self, deploy_model: "DeployModel"):
            """
            :param deploy_model: Parent DeployModel widget.
            :type deploy_model: DeployModel, optional
            """
            self.api = deploy_model.api
            self.team_id = deploy_model.team_id
            self._cache = deploy_model._cache
            self.deploy_model = deploy_model
            self._model_api = None
            self._lock = threading.Lock()
            self._layout = self._create_layout()

        @property
        def layout(self) -> ExperimentSelector:
            return self._layout

        def _create_layout(self) -> Container:
            self.experiment_table = ExperimentSelector(
                api=self.api,
                team_id=self.team_id,
            )

            @self.experiment_table.checkpoint_changed
            def _checkpoint_changed(row: ExperimentSelector.ModelRow, checkpoint_value: str):
                print(f"Checkpoint changed for {row._experiment_info.task_id}: {checkpoint_value}")

            threading.Thread(target=self.refresh_experiments, daemon=True).start()

            return self.experiment_table

        def refresh_experiments(self):
            with self._lock:
                self.experiment_table.loading = True
                frameworks = self.deploy_model.get_frameworks()
                experiment_infos = []
                for framework_name in frameworks:
                    experiment_infos.extend(
                        get_experiment_infos(self.api, self.team_id, framework_name=framework_name)
                    )

                self.experiment_table.set_experiment_infos(experiment_infos)
                self.experiment_table.loading = False

        def get_deploy_parameters(self) -> Dict[str, Any]:
            experiment_info = self.experiment_table.get_selected_experiment_info()
            return {
                "experiment_info": (experiment_info.to_json() if experiment_info else None),
            }

        def deploy(self, agent_id: int) -> ModelAPI:
            deploy_parameters = self.get_deploy_parameters()
            logger.info(f"Deploying custom model with parameters:", extra=deploy_parameters)
            experiment_info = deploy_parameters["experiment_info"]
            experiment_info = ExperimentInfo(**experiment_info)  # pylint: disable=not-a-mapping
            task_info = self.api.nn._deploy_api.deploy_custom_model_from_experiment_info(
                agent_id=agent_id,
                experiment_info=experiment_info,
                log_level="debug",
            )
            model_api = ModelAPI(api=self.api, task_id=task_info["id"])
            return model_api

        def load_from_json(self, data: Dict):
            if "experiment_info" in data:
                with self._lock:
                    experiment_info_json = data["experiment_info"]
                    experiment_info = ExperimentInfo(**experiment_info_json)  # pylint: disable=not-a-mapping
                    self.experiment_table.set_selected_row_by_experiment_info(experiment_info)
            elif "train_task_id" in data:
                with self._lock:
                    task_id = data["train_task_id"]
                    self.experiment_table.set_selected_row_by_task_id(task_id)
            else:
                raise ValueError("Invalid data format for loading custom model.")

    class MODE:
        """Supported deployment modes for the widget."""
        CONNECT = "connect"
        PRETRAINED = "pretrained"
        CUSTOM = "custom"

    MODES = [str(MODE.CONNECT), str(MODE.PRETRAINED), str(MODE.CUSTOM)]
    MODE_TO_CLASS = {
        str(MODE.CONNECT): Connect,
        str(MODE.CUSTOM): Custom,
        str(MODE.PRETRAINED): Pretrained,
    }

    def __init__(
        self,
        api: Api = None,
        team_id: int = None,
        modes: List[Literal["connect", "pretrained", "custom"]] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param api: Supervisely API instance. Defaults to Api.from_env().
        :type api: :class:`~supervisely.api.api.Api`, optional
        :param team_id: Team ID. Defaults to env.team_id().
        :type team_id: int, optional
        :param modes: List of deployment modes: "connect", "pretrained", "custom".
        :type modes: List[Literal["connect", "pretrained", "custom"]], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self.modes: Dict[str, DeployModel.DeployMode] = {}
        if modes is None:
            modes = self.MODES.copy()
        self._validate_modes(modes)
        if api is None:
            api = Api()
        self.api = api
        if team_id is None:
            team_id = env.team_id()
        self.team_id = team_id
        self._cache = {}

        self.modes_labels = {
            self.MODE.CONNECT: "Connect",
            self.MODE.PRETRAINED: "Pretrained",
            self.MODE.CUSTOM: "Custom",
        }
        self.modes_descriptions = {
            self.MODE.CONNECT: "Connect to an already deployed model",
            self.MODE.PRETRAINED: "Deploy a pretrained model from the ecosystem",
            self.MODE.CUSTOM: "Deploy a custom model from your experiments",
        }

        # GUI
        self.layout: Widget = None
        self._init_gui(modes)

        self.model_api: ModelAPI = None

        super().__init__(widget_id=widget_id)

    def _validate_modes(self, modes) -> None:
        if len(modes) < 1 or len(modes) > len(self.MODES):
            raise ValueError(
                f"Modes must be a list containing 1 to {len(self.MODES)} of the following: {', '.join(self.MODES)}."
            )
        for mode in modes:
            if mode not in self.MODES:
                raise ValueError(f"Invalid mode '{mode}'. Valid modes are {', '.join(self.MODES)}.")

    def get_modules(self) -> List[ModuleInfo]:
        modules = self._cache.setdefault("modules", [])
        if len(modules) > 0:
            return modules
        modules = self.api.app.get_list_ecosystem_modules(
            categories=["serve", "images"], categories_operation="and"
        )
        modules = [
            module
            for module in modules
            if any([cat for cat in module["config"]["categories"] if cat.startswith("framework:")])
        ]
        modules = [ModuleInfo.from_json(module) for module in modules]
        self._cache["modules"] = modules
        return modules

    def get_frameworks(self) -> List[str]:
        if len(self._cache.get("frameworks", [])) > 0:
            return self._cache["frameworks"]

        modules = self._cache.setdefault("modules", [])
        if len(modules) == 0:
            modules = self.get_modules()
        frameworks = [cat for module in modules for cat in module.config.get("categories", [])]
        frameworks = [
            cat[len("framework:") :] for cat in frameworks if cat.startswith("framework:")
        ]
        self._cache["frameworks"] = frameworks
        return frameworks

    def _init_modes(self, modes: str) -> None:
        for mode in modes:
            self.modes[mode] = self.MODE_TO_CLASS[mode](self)

    def _create_task_link(self, task_id: int) -> str:
        return f"{self.api.server_address}/apps/sessions/{task_id}"

    def _get_inference_settings_by_module(self, module: Dict) -> str:
        config = module["config"]
        inference_settings_path = config.get("files", {}).get("inference_settings", None)
        if inference_settings_path is None:
            raise ValueError(
                f"No inference settings file found for framework app {module['meta']['app']['name']}."
            )
        save_path = tempfile.mktemp(suffix=".yaml")
        self.api.app.download_git_file(
            module_id=module["id"],
            file_path=inference_settings_path,
            save_path=save_path,
        )
        inference_settings = Path(save_path).read_text()
        return inference_settings

    def _get_inference_settings_for_framework(self, framework: str) -> str:
        inference_settings_cache = self._cache.setdefault("inference_settings", {})
        if framework not in inference_settings_cache:
            module = self.api.nn._deploy_api.find_serving_app_by_framework(framework)
            if module is None:
                raise ValueError(f"No serving app found for framework {framework}.")
            config = module["config"]
            inference_settings_path = config.get("files", {}).get("inference_settings", None)
            if inference_settings_path is None:
                raise ValueError(f"No inference settings file found for framework {framework}.")
            save_path = tempfile.mktemp(suffix=".yaml")
            self.api.app.download_git_file(
                module_id=module["id"],
                file_path=inference_settings_path,
                save_path=save_path,
            )
            inference_settings = Path(save_path).read_text()
            inference_settings_cache[framework] = inference_settings
        return inference_settings_cache[framework]

    def _framework_from_task_info(self, task_info: Dict) -> str:
        module_id = task_info["meta"]["app"]["moduleId"]
        module = None
        for m in self.get_modules():
            if m.id == module_id:
                module = m
        if module is None:
            module = self.api.app.get_ecosystem_module_info(module_id=module_id)
            self._cache.setdefault("modules", []).append(module)
        for cat in module.config["categories"]:
            if cat.startswith("framework:"):
                return cat[len("framework:") :]
        return "unknown"

    def _init_gui(self, modes: List[str]) -> None:
        self.status = Text("Deploying model...", status="info")
        self.session_text_1 = Text(
            "",
            "text",
        )
        self.session_text_2 = Text(
            "",
            "text",
            font_size=13,
        )
        self.sesson_link = Container(
            [
                self.session_text_1,
                self.session_text_2,
            ],
            gap=0,
            style="padding-left: 10px",
        )
        self.status.hide()
        self.sesson_link.hide()

        self.select_agent = AgentSelector(self.team_id)
        self.select_agent_field = Field(content=self.select_agent, title="Select Agent")

        self._create_model_info_widget()

        self.deploy_button = Button("Deploy", icon="zmdi zmdi-play")
        self.connect_button = Button("Connect", icon="zmdi zmdi-link")
        self.stop_button = Button("Stop", icon="zmdi zmdi-stop", button_type="danger")
        self.stop_button.hide()
        self.disconnect_button = Button("Disconnect", icon="zmdi zmdi-close", button_type="warning")
        self.disconnect_button.hide()
        self.deploy_stop_buttons = Flexbox(
            widgets=[self.deploy_button, self.stop_button, self.disconnect_button],
            gap=10,
        )
        self.connect_stop_buttons = Flexbox(
            widgets=[self.connect_button, self.stop_button, self.disconnect_button],
            gap=10,
        )

        self._init_modes(modes)
        _labels = []
        _descriptions = []
        _contents = []
        self.statuses_widgets = Container(
            widgets=[
                self.sesson_link,
                self._model_info_container,
            ],
            gap=20,
        )
        self.statuses_widgets.hide()
        for mode_name, mode in self.modes.items():
            label = self.modes_labels[mode_name]
            description = self.modes_descriptions[mode_name]
            if mode_name == str(self.MODE.CONNECT):
                widgets = [
                    mode.layout,
                    self.status,
                    self.statuses_widgets,
                    self.connect_stop_buttons,
                ]
            else:
                widgets = [
                    mode.layout,
                    self.select_agent_field,
                    self.status,
                    self.statuses_widgets,
                    self.deploy_stop_buttons,
                ]

            content = Container(widgets=widgets, gap=20)
            _labels.append(label)
            _descriptions.append(description)
            _contents.append(content)

        self.tabs = RadioTabs(titles=_labels, descriptions=_descriptions, contents=_contents)
        if len(self.modes) == 1:
            self.layout = _contents[0]
        else:
            self.layout = self.tabs

        @self.deploy_button.click
        def _deploy_button_clicked():
            self._deploy()

        @self.stop_button.click
        def _stop_button_clicked():
            self.stop()

        @self.connect_button.click
        def _connect_button_clicked():
            self._connect()

        @self.disconnect_button.click
        def _disconnect_button_clicked():
            self.disconnect()

        @self.tabs.value_changed
        def _active_tab_changed(tab_name: str):
            self.set_model_message_by_tab(tab_name)

    def set_model_status(
        self,
        status: Literal["deployed", "stopped", "deploying", "connecting", "error", "hide"],
        extra_text: str = None,
    ) -> None:
        if status == "hide":
            self.status.hide()
            return
        status_args = {
            "deployed": {"text": "Model deployed successfully!", "status": "success"},
            "stopped": {"text": "Model stopped", "status": "info"},
            "deploying": {"text": "Deploying model...", "status": "info"},
            "connecting": {"text": "Connecting to model...", "status": "info"},
            "connected": {"text": "Model connected successfully!", "status": "success"},
            "error": {
                "text": "Error occurred during model deployment.",
                "status": "error",
            },
        }
        args = status_args[status]
        if extra_text:
            args["text"] += f" {extra_text}"
        self.status.set(**args)
        self.status.show()

    def set_session_info(self, task_info: Dict):
        if task_info is None:
            self.sesson_link.hide()
            return
        task_id = task_info["id"]
        task_link = self._create_task_link(task_id)
        task_date = task_info["startedAt"]
        task_date = datetime.datetime.fromisoformat(task_date.replace("Z", "+00:00"))
        task_date = task_date.strftime("%Y-%m-%d %H:%M:%S")
        task_name = task_info["meta"]["app"]["name"]
        self.session_text_1.text = f"<i class='zmdi zmdi-link' style='color: #7f858e'></i> <a href='{task_link}' target='_blank'>{task_name}: {task_id}</a>"
        self.session_text_2.text = f"<span class='field-description text-muted' style='color: #7f858e'>{task_date} (UTC)</span>"
        self.sesson_link.show()

    def disable_modes(self) -> None:
        for mode_name, mode in self.modes.items():
            mode.layout.disable()
            label = self.modes_labels[mode_name]
            self.tabs.disable_tab(label)
        self.select_agent.disable()

    def enable_modes(self) -> None:
        for mode_name, mode in self.modes.items():
            mode.layout.enable()
            label = self.modes_labels[mode_name]
            self.tabs.enable_tab(label)
        self.select_agent.enable()

    def show_deploy_button(self) -> None:
        self.stop_button.hide()
        self.disconnect_button.hide()
        self.connect_button.show()
        self.deploy_button.show()

    def show_stop(self) -> None:
        self.connect_button.hide()
        self.deploy_button.hide()
        self.stop_button.show()
        self.disconnect_button.show()

    def _connect(self) -> None:
        self.set_model_status("connecting")
        self.set_session_info(None)
        try:
            self.disable_modes()
            model_api = self.deploy()
            task_info = self.api.task.get_info_by_id(model_api.task_id)
            model_info = model_api.get_info()
            model_name = model_info["model_name"]
            framework = self._framework_from_task_info(task_info)
            logger.info(
                f"Model {framework}: {model_name} deployed with session ID {model_api.task_id}."
            )
            self.model_api = model_api
            self.statuses_widgets.show()
            self.set_model_status("connected")
            self.set_session_info(task_info)
            self.set_model_info(model_api.task_id)
            self.show_stop()
        except Exception as e:
            logger.error(f"Failed to deploy model: {e}", exc_info=True)
            self.set_model_status("error", str(e))
            self.set_session_info(None)
            self.enable_modes()
            self.reset_model_info()
            self.show_deploy_button()

    def _deploy(self) -> None:
        self.set_model_status("deploying")
        self.set_session_info(None)
        try:
            self.disable_modes()
            model_api = self.deploy()
            task_info = self.api.task.get_info_by_id(model_api.task_id)
            model_info = model_api.get_info()
            model_name = model_info["model_name"]
            framework = self._framework_from_task_info(task_info)
            logger.info(
                f"Model {framework}: {model_name} deployed with session ID {model_api.task_id}."
            )
            self.model_api = model_api
            self.set_model_status("deployed")
            self.set_session_info(task_info)
            self.set_model_info(model_api.task_id)
            self.show_stop()
            self.statuses_widgets.show()
        except Exception as e:
            logger.error(f"Failed to deploy model: {e}", exc_info=True)
            self.set_model_status("error", str(e))
            self.set_session_info(None)
            self.reset_model_info()
            self.show_deploy_button()
            self.statuses_widgets.hide()
            self.enable_modes()
        else:
            if str(self.MODE.CONNECT) in self.modes:
                self.modes[str(self.MODE.CONNECT)]._update_sessions()

    def deploy(self) -> ModelAPI:
        mode_label = self.tabs.get_active_tab()
        mode = None
        for mode, label in self.modes_labels.items():
            if label == mode_label:
                break
        agent_id = self.select_agent.get_value()
        self.model_api = self.modes[mode].deploy(agent_id=agent_id)
        return self.model_api

    def stop(self) -> None:
        if self.model_api is None:
            return
        logger.info("Stopping model...")
        self.model_api.shutdown()
        self.model_api = None
        self.set_model_status("stopped")
        self.enable_modes()
        self.reset_model_info()
        self.show_deploy_button()
        self.statuses_widgets.hide()
        if str(self.MODE.CONNECT) in self.modes:
            self.modes[str(self.MODE.CONNECT)]._update_sessions()

    def disconnect(self) -> None:
        if self.model_api is None:
            return
        self.model_api = None
        self.set_model_status("hide")
        self.set_session_info(None)
        self.reset_model_info()
        self.show_deploy_button()
        self.statuses_widgets.hide()
        self.enable_modes()

    def load_from_json(self, data: Dict[str, Any]) -> None:
        """
        Load widget state from JSON data.
        :param data: Dictionary with widget data.
        """
        if not data:
            return
        mode = data["mode"]
        label = self.modes_labels[mode]
        self.tabs.set_active_tab(label)
        agent_id = data.get("agent_id", None)
        if agent_id is not None:
            self.select_agent.set_value(agent_id)
        self.modes[mode].load_from_json(data)

    def get_deploy_parameters(self) -> Dict[str, Any]:
        mode_label = self.tabs.get_active_tab()
        mode = None
        for mode, label in self.modes_labels.items():
            if label == mode_label:
                break
        agent_id = self.select_agent.get_value()
        parameters = {"mode": mode, "agent_id": agent_id}
        parameters.update(self.modes[mode].get_deploy_parameters())
        return parameters

    def get_json_data(self) -> Dict[str, Any]:
        return {}

    def get_json_state(self) -> Dict[str, Any]:
        return {}

    def to_html(self):
        return self.layout.to_html()

    # Model Info
    def _create_model_info_widget(self):
        self._model_info_widget = ModelInfo()
        self._model_info_widget_field = Field(
            self._model_info_widget,
            title="Model Info",
            description="Information about the deployed model",
        )
        self._model_info_message = Text("Connect to model to see the session information.")
        self._model_info_container = Container(
            [self._model_info_widget_field, self._model_info_message], gap=0
        )
        self._model_info_widget_field.hide()

        self._model_info_container.hide()

    def set_model_info(self, session_id):
        self._model_info_widget.set_session_id(session_id)

        self._model_info_message.hide()
        self._model_info_widget_field.show()
        self._model_info_container.show()

    def reset_model_info(self):
        self._model_info_container.hide()
        self._model_info_widget_field.hide()
        self._model_info_message.show()

    def set_model_message_by_tab(self, tab_name: str):
        if tab_name == self.modes_labels[str(self.MODE.CONNECT)]:
            self._model_info_message.set(
                "Connect to model to see the session information.", status="text"
            )
        else:
            self._model_info_message.set(
                "Deploy model to see the session information.", status="text"
            )
        self._model_info_widget_field.hide()

    # ------------------------------------------------------------ #



from supervisely.nn.artifacts.detectron2 import Detectron2
from supervisely.nn.artifacts.hrda import HRDA
from supervisely.nn.artifacts.mmclassification import MMClassification, MMPretrain
from supervisely.nn.artifacts.mmdetection import MMDetection, MMDetection3
from supervisely.nn.artifacts.mmsegmentation import MMSegmentation
from supervisely.nn.artifacts.ritm import RITM
from supervisely.nn.artifacts.rtdetr import RTDETR
from supervisely.nn.artifacts.unet import UNet
from supervisely.nn.artifacts.yolov5 import YOLOv5, YOLOv5v2
from supervisely.nn.artifacts.yolov8 import YOLOv8



from typing import Dict, Optional

from supervisely.nn.benchmark.base_visualizer import BaseVisMetric
from supervisely.nn.benchmark.semantic_segmentation.evaluator import (
    SemanticSegmentationEvalResult,
)


class SemanticSegmVisMetric(BaseVisMetric):
    """Base visualization metric for semantic segmentation reports with segmentation-specific click data."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_result: SemanticSegmentationEvalResult

    def get_click_data(self) -> Optional[Dict]:
        if not self.clickable:
            return

        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        for key, v in self.eval_result.images_by_class.items():
            res["clickData"][key] = {}

            title = f"Class {key}: {len(v)} image{'s' if len(v) > 1 else ''}"
            res["clickData"][key]["title"] = title
            img_ids = [
                self.eval_result.matched_pair_data[gt_img_id].pred_image_info.id for gt_img_id in v
            ]
            res["clickData"][key]["imagesIds"] = img_ids

        return res

    def get_diff_data(self) -> Dict:
        res = {}

        res["layoutTemplate"] = [
            {"columnTitle": "Original Image"},
            {"columnTitle": "Ground Truth Masks"},
            {"columnTitle": "Predicted Masks"},
        ]

        click_data = res.setdefault("clickData", {})

        for pairs_data in self.eval_result.matched_pair_data.values():
            gt = pairs_data.gt_image_info
            pred = pairs_data.pred_image_info
            diff = pairs_data.diff_image_info
            assert gt.name == pred.name == diff.name
            key = click_data.setdefault(str(pred.id), {})
            key["imagesIds"] = [diff.id, gt.id, pred.id]
            key["title"] = f"Image: {pred.name}"
        return res



from typing import List, Optional

from supervisely.api.user_api import UserInfo
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget


class MembersListSelector(Widget):
    """Widget to select one or multiple team members from a provided list of users."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        CHECKBOX_CHANGED = "checkbox_cb"

    def __init__(
        self,
        users: Optional[List[UserInfo]] = [],
        multiple: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """
        :param users: List of UserInfo to select from.
        :type users: List[UserInfo], optional
        :param multiple: If True, allow multiple selection.
        :type multiple: bool, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._users = users
        self._multiple = multiple
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "users": [
                {
                    "login": f"<b>{user.login}</b>",
                    "role": user.role,
                }
                for user in self._users
            ]
        }

    def get_json_state(self):
        return {"selected": [False for _ in self._users]}

    def set(self, users: List[UserInfo]):
        selected_members = [name for name in self.get_selected_members()]
        self._users = users
        StateJson()[self.widget_id]["selected"] = [
            user.login in selected_members for user in self._users
        ]
        self.update_data()
        StateJson().send_changes()

    def get_selected_members(self):
        selected = StateJson()[self.widget_id]["selected"]
        return [user for user, is_selected in zip(self._users, selected) if is_selected]

    def select_all(self):
        StateJson()[self.widget_id]["selected"] = [True for _ in self._users]
        StateJson().send_changes()

    def deselect_all(self):
        StateJson()[self.widget_id]["selected"] = [False for _ in self._users]
        StateJson().send_changes()

    def select(self, names: List[str]):
        selected = [user.login in names for user in self._users]
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def deselect(self, names: List[str]):
        selected = StateJson()[self.widget_id]["selected"]
        for idx, user in enumerate(self._users):
            if user.name in names:
                selected[idx] = False
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def set_multiple(self, value: bool):
        self._multiple = value

    def get_all_members(self):
        return self._users

    def selection_changed(self, func):
        route_path = self.get_route_path(MembersListSelector.Routes.CHECKBOX_CHANGED)
        server = self._sly_app.get_server()
        self._checkboxes_handled = True

        @server.post(route_path)
        def _click():
            selected = self.get_selected_members()
            func(selected)

        return _click



from typing import List

import supervisely.nn.benchmark.comparison.semantic_segmentation.text_templates as texts
from supervisely.nn.benchmark.comparison.base_visualizer import BaseComparisonVisualizer
from supervisely.nn.benchmark.comparison.semantic_segmentation.vis_metrics import (
    ClasswiseErrorAnalysis,
    ExplorePredictions,
    FrequentlyConfused,
    IntersectionErrorOverUnion,
    Overview,
    RenormalizedErrorOverUnion,
    Speedtest,
)
from supervisely.nn.benchmark.semantic_segmentation.evaluator import (
    SemanticSegmentationEvalResult,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ContainerWidget,
    MarkdownWidget,
    SidebarWidget,
)


class SemanticSegmentationComparisonVisualizer(BaseComparisonVisualizer):
    """Visualizer for semantic segmentation model comparison reports."""

    vis_texts = texts
    ann_opacity = 0.7

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_results: List[SemanticSegmentationEvalResult]

    def _create_widgets(self):
        # Modal Gellery
        self.diff_modal = self._create_diff_modal_table()
        self.explore_modal = self._create_explore_modal_table(
            click_gallery_id=self.diff_modal.id, hover_text="Compare with GT"
        )

        # Notifcation
        self.clickable_label = self._create_clickable_label()

        # Speedtest init here for overview
        speedtest = Speedtest(self.vis_texts, self.comparison.eval_results)

        # Overview
        overview = Overview(self.vis_texts, self.comparison.eval_results)
        overview.team_id = self.comparison.team_id
        self.header = self._create_header()
        self.overviews = self._create_overviews(overview, grid_cols=2)
        self.overview_md = overview.overview_md
        self.key_metrics_md = self._create_key_metrics()
        self.key_metrics_table = overview.get_table_widget(
            latency=speedtest.latency, fps=speedtest.fps
        )
        self.overview_chart = overview.chart_widget

        # Explore Predictions
        columns_number = len(self.comparison.eval_results) + 1  # +1 for GT
        self.explore_predictions_modal_gallery = self._create_explore_modal_table(columns_number)
        explore_predictions = ExplorePredictions(
            self.vis_texts,
            self.comparison.eval_results,
            explore_modal_table=self.explore_predictions_modal_gallery,
        )
        self.explore_predictions_md = explore_predictions.difference_predictions_md
        self.explore_predictions_gallery = explore_predictions.explore_gallery

        # IntersectionErrorOverUnion
        iou_eou = IntersectionErrorOverUnion(self.vis_texts, self.comparison.eval_results)
        self.iou_eou_md = iou_eou.md
        self.iou_eou_chart = iou_eou.chart

        # RenormalizedErrorOverUnion
        reou = RenormalizedErrorOverUnion(self.vis_texts, self.comparison.eval_results)
        self.reou_md = reou.md
        self.reou_chart = reou.chart

        # ClasswiseErrorAnalysis
        classwise_ea = ClasswiseErrorAnalysis(self.vis_texts, self.comparison.eval_results)
        self.classwise_ea_md = classwise_ea.md
        self.classwise_ea_chart = classwise_ea.chart

        # FrequentlyConfused
        frequently_confused = FrequentlyConfused(self.vis_texts, self.comparison.eval_results)
        self.frequently_confused_md = frequently_confused.md
        self.frequently_confused_chart = frequently_confused.chart

        # # SpeedTest
        self.speedtest_present = not speedtest.is_empty()
        self.speedtest_multiple_batch_sizes = False

        if self.speedtest_present:
            self.speedtest_md_intro = speedtest.md_intro
            self.speedtest_intro_table = speedtest.intro_table
            self.speed_inference_time_md = speedtest.inference_time_md
            self.speed_inference_time_table = speedtest.inference_time_table
            self.speed_fps_md = speedtest.fps_md
            self.speed_fps_table = speedtest.fps_table
            self.speedtest_multiple_batch_sizes = speedtest.multiple_batche_sizes()
            if self.speedtest_multiple_batch_sizes:
                self.speed_batch_inference_md = speedtest.batch_inference_md
                self.speed_chart = speedtest.chart

    def _create_layout(self):
        is_anchors_widgets = [
            # Overview
            (0, self.header),
            (1, self.overview_md),
            (0, self.overviews),
            (1, self.key_metrics_md),
            (0, self.key_metrics_table),
            (0, self.overview_chart),
            # Explore Predictions
            (1, self.explore_predictions_md),
            (0, self.explore_predictions_gallery),
            # IntersectionErrorOverUnion
            (1, self.iou_eou_md),
            (0, self.iou_eou_chart),
            # RenormalizedErrorOverUnion
            (1, self.reou_md),
            (0, self.reou_chart),
            # ClasswiseErrorAnalysis
            (1, self.classwise_ea_md),
            (0, self.classwise_ea_chart),
            # FrequentlyConfused
            (1, self.frequently_confused_md),
            (0, self.frequently_confused_chart),
        ]
        if self.speedtest_present:
            is_anchors_widgets.extend(
                [
                    # SpeedTest
                    (1, self.speedtest_md_intro),
                    (0, self.speedtest_intro_table),
                    (0, self.speed_inference_time_md),
                    (0, self.speed_inference_time_table),
                    (0, self.speed_fps_md),
                    (0, self.speed_fps_table),
                ]
            )
            if self.speedtest_multiple_batch_sizes:
                is_anchors_widgets.append((0, self.speed_batch_inference_md))
                is_anchors_widgets.append((0, self.speed_chart))
        anchors = []
        for is_anchor, widget in is_anchors_widgets:
            if is_anchor:
                anchors.append(widget.id)

        sidebar = SidebarWidget(widgets=[i[1] for i in is_anchors_widgets], anchors=anchors)
        layout = ContainerWidget(
            widgets=[sidebar, self.explore_modal, self.explore_predictions_modal_gallery],
            name="main_container",
        )
        return layout

    def _create_key_metrics(self) -> MarkdownWidget:
        return MarkdownWidget(
            "markdown_key_metrics", "Key Metrics", text=self.vis_texts.markdown_key_metrics
        )



# coding: utf-8

# docs
from __future__ import annotations

from copy import deepcopy
from typing import Dict, Iterable, List, Optional, Tuple, Union

import cv2
import numpy as np

from supervisely.geometry.constants import (
    EXTERIOR,
    GEOMETRY_SHAPE,
    GEOMETRY_TYPE,
    INTERIOR,
    POINTS,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list
from supervisely.geometry.rectangle import Rectangle


class VectorGeometry(Geometry):
    """Base class for vector geometries (polygon, polyline) defined by exterior/interior contours. Immutable."""

    def __init__(
        self,
        exterior: Union[List[PointLocation], List[List[int, int]], List[Tuple[int, int]]],
        interior: Union[
            List[List[PointLocation]],
            List[List[List[int, int]]],
            List[List[Tuple[int, int]]],
        ] = [],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        VectorGeometry is a base class of geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.vector_geometry.VectorGeometry` object is immutable.

        :param exterior: Exterior coordinates, object contour is defined with these points (used for :class:`~supervisely.geometry.polygon.Polygon`).
        :type exterior: List[:class:`~supervisely.geometry.point_location.PointLocation`], List[List[int, int]], List[Tuple[int, int]
        :param interior: Interior coordinates, object holes is defined with these points (used for :class:`~supervisely.geometry.polygon.Polygon`).
        :type interior: List[List[:class:`~supervisely.geometry.point_location.PointLocation`]], List[List[List[int, int]]], List[List[Tuple[int, int]]]
        :param sly_id: VectorGeometry ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of :class:`~supervisely.project.obj_class.ObjClass` to which VectorGeometry belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created VectorGeometry.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VectorGeometry was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VectorGeometry was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :raises TypeError: if exterior or interior parameters are not a list of :class:`~supervisely.geometry.point_location.PointLocation` objects

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                exterior = [sly.PointLocation(730, 2104), sly.PointLocation(2479, 402), sly.PointLocation(3746, 1646)]
                # or exterior = [[730, 2104], [2479, 402], [3746, 1646]]
                # or exterior = [(730, 2104), (2479, 402), (3746, 1646)]
                interior = [[sly.PointLocation(1907, 1255), sly.PointLocation(2468, 875), sly.PointLocation(2679, 1577)]]
                # or interior = [[[730, 2104], [2479, 402], [3746, 1646]]]
                # or interior = [[(730, 2104), (2479, 402), (3746, 1646)]]

                figure = sly.Polygon(exterior, interior)
        """
        result_exterior = []
        if not isinstance(exterior, list):
            raise TypeError('Argument "exterior" must be a list of coordinates')
        for p in exterior:
            if isinstance(p, PointLocation):
                result_exterior.append(p)
            elif isinstance(p, tuple) and len(p) == 2:
                result_exterior.append(PointLocation(p[0], p[1]))
            elif isinstance(p, list) and len(p) == 2:
                result_exterior.append(PointLocation(p[0], p[1]))
            else:
                raise TypeError(
                    'Type of items (coordinates) in list "exterior" have to be tuple(int, int) or list[int, int] or PointLocation(row, col)'
                )

        result_interior = []
        if not isinstance(interior, list):
            raise TypeError('Argument "interior" must be a list of lists with coordinates')
        for coords in interior:
            if not isinstance(interior, list):
                raise TypeError('"interior" coords must be a list of coordinates')
            p_coords = []
            for p in coords:
                if isinstance(p, PointLocation):
                    p_coords.append(p)
                elif isinstance(p, tuple) and len(p) == 2:
                    p_coords.append(PointLocation(p[0], p[1]))
                elif isinstance(p, list) and len(p) == 2:
                    p_coords.append(PointLocation(p[0], p[1]))
                else:
                    raise TypeError(
                        'Type of items (coordinates) in list "interior" have to be tuple(int, int) or list[int, int] or PointLocation(row, col)'
                    )
            result_interior.append(p_coords)

        self._exterior = deepcopy(result_exterior)
        self._interior = deepcopy(result_interior)
        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def to_json(self) -> Dict:
        """
        Convert the VectorGeometry to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Json format as a dict
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                figure_json = figure.to_json()
                print(figure_json)
                # Output: {
                #    "points": {
                #        "exterior": [
                #            [2104, 730],
                #            [402, 2479],
                #            [1646, 3746]
                #        ],
                #        "interior": [
                #            [
                #                [1255, 1907],
                #                [875, 2468],
                #                [1577, 2679]
                #            ]
                #        ]
                #    }
                # }
        """
        packed_obj = {
            POINTS: {
                EXTERIOR: points_to_row_col_list(self._exterior, flip_row_col_order=True),
                INTERIOR: [
                    points_to_row_col_list(i, flip_row_col_order=True) for i in self._interior
                ],
            },
            GEOMETRY_SHAPE: self.geometry_name(),
            GEOMETRY_TYPE: self.geometry_name(),
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @property
    def exterior(self) -> List[PointLocation]:
        """
        VectorGeometry exterior points.

        :returns: VectorGeometry exterior points
        :rtype: List[:class:`~supervisely.geometry.point_location.PointLocation`]

        :Usage Example:

            .. code-block:: python

                exterior = figure.exterior
        """
        return deepcopy(self._exterior)

    @property
    def exterior_np(self) -> np.ndarray:
        """
        Converts exterior attribute of VectorGeometry to numpy array.

        :returns: Numpy array
        :rtype: np.ndarray

        :Usage Example:

            .. code-block:: python

                print(figure.exterior_np)
                # Output:
                # [[ 730 2104]
                #  [2479  402]
                #  [3746 1646]]
        """
        return np.array(points_to_row_col_list(self._exterior), dtype=np.int64)

    @property
    def interior(self) -> List[List[PointLocation]]:
        """
        VectorGeometry interior points.

        :returns: VectorGeometry interior points
        :rtype: List[List[:class:`~supervisely.geometry.point_location.PointLocation`]]

        :Usage Example:

            .. code-block:: python

                interior = figure.interior
        """
        return deepcopy(self._interior)

    @property
    def interior_np(self):
        """
        Converts interior attribute of VectorGeometry to numpy array.

        :returns: Numpy array
        :rtype: List[np.ndarray]

        :Usage Example:

            .. code-block:: python

                print(figure.interior_np)
                # Output:
                # [array([[1907, 1255],
                #        [2468,  875],
                #        [2679, 1577]])]
        """
        return [np.array(points_to_row_col_list(i), dtype=np.int64) for i in self._interior]

    def _transform(self, transform_fn):
        """ """
        result = deepcopy(self)
        result._exterior = [transform_fn(p) for p in self._exterior]
        result._interior = [[transform_fn(p) for p in i] for i in self._interior]
        return result

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> VectorGeometry:
        """
        Resizes current VectorGeometry.

        :param in_size: Input image size (height, width) to which belongs VectorGeometry.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs VectorGeometry.
        :type out_size: Tuple[int, int]
        :returns: Resized VectorGeometry.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                in_height, in_width = 300, 400
                out_height, out_width = 600, 800
                resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor: float) -> VectorGeometry:
        """
        Scales current VectorGeometry.

        :param factor: Scale parameter.
        :type factor: float
        :returns: Scaled VectorGeometry.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                scale_figure = figure.scale(0.75)
        """
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow: int, dcol: int) -> VectorGeometry:
        """
        Translates current VectorGeometry.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :returns: Translated VectorGeometry.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                translate_figure = figure.translate(150, 250)
        """
        return self._transform(lambda p: p.translate(drow, dcol))

    def rotate(self, rotator: ImageRotator) -> VectorGeometry:
        """
        Rotates current VectorGeometry.

        :param rotator: ImageRotator for rotate.
        :type rotator: :class:`~supervisely.geometry.image_rotator.ImageRotator`
        :returns: Rotated VectorGeometry.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                from supervisely.geometry.image_rotator import ImageRotator

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                height, width = 300, 400
                rotator = ImageRotator((height, width), 25)
                rotate_figure = figure.rotate(rotator)

        """
        return self._transform(lambda p: p.rotate(rotator))

    def fliplr(self, img_size: Tuple[int, int]) -> VectorGeometry:
        """
        Flips current VectorGeometry in horizontal.

        :param img_size: Input image size (height, width) to which belongs VectorGeometry.
        :type img_size: Tuple[int, int]
        :returns: Flipped VectorGeometry in horizontal.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                height, width = 300, 400
                fliplr_figure = figure.fliplr((height, width))
        """
        return self._transform(lambda p: p.fliplr(img_size))

    def flipud(self, img_size: Tuple[int, int]) -> VectorGeometry:
        """
        Flips current VectorGeometry in vertical.

        :param img_size: Input image size (height, width) to which belongs VectorGeometry.
        :type img_size: Tuple[int, int]
        :returns: Flipped VectorGeometry in vertical.
        :rtype: :class:`~supervisely.geometry.vector_geometry.VectorGeometry`

        :Usage Example:

            .. code-block:: python

                # Remember that VectorGeometry class object is immutable, and we need to assign new instance of VectorGeometry to a new variable
                height, width = 300, 400
                flipud_figure = figure.flipud((height, width))
        """
        return self._transform(lambda p: p.flipud(img_size))

    def to_bbox(self) -> Rectangle:
        """
        Creates Rectangle object from current VectorGeometry.

        :returns: Rectangle from VectorGeometry.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                rectangle = figure.to_bbox()
        """
        exterior_np = self.exterior_np
        rows, cols = exterior_np[:, 0], exterior_np[:, 1]
        return Rectangle(
            top=round(min(rows).item()),
            left=round(min(cols).item()),
            bottom=round(max(rows).item()),
            right=round(max(cols).item()),
        )

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: used only in Polyline and Point.
        """
        self._draw_contour_impl(bitmap, color, thickness, config=config)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        """
        raise NotImplementedError()

    @staticmethod
    def _approx_ring_dp(ring, epsilon, closed):
        """ """
        new_ring = cv2.approxPolyDP(ring.astype(np.int32), epsilon, closed)
        new_ring = np.squeeze(new_ring, 1)
        if len(new_ring) < 3 and closed:
            new_ring = ring.astype(np.int32)
        return new_ring

    def approx_dp(self, epsilon):
        """ """
        raise NotImplementedError()



from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class IntersectionErrorOverUnion(BaseVisMetrics):
    """IoU and error-over-union charts for segmentation model comparison."""


    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "intersection_error_over_union",
            "Intersection & Error Over Union",
            text=self.vis_texts.markdown_iou,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("intersection_error_over_union", self.get_figure())

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error
        from plotly.subplots import make_subplots  # pylint: disable=import-error

        labels = ["mIoU", "mBoundaryEoU", "mExtentEoU", "mSegmentEoU"]

        length = len(self.eval_results)
        cols = 3 if length > 2 else 2
        cols = 4 if length % 4 == 0 else cols
        rows = length // cols + (1 if length % cols != 0 else 0)

        fig = make_subplots(rows=rows, cols=cols, specs=[[{"type": "domain"}] * cols] * rows)

        annotations = []
        for idx, eval_result in enumerate(self.eval_results, start=1):
            col = idx % cols + (cols if idx % cols == 0 else 0)
            row = idx // cols + (1 if idx % cols != 0 else 0)

            fig.add_trace(
                go.Pie(
                    labels=labels,
                    values=[
                        eval_result.mp.iou,
                        eval_result.mp.boundary_eou,
                        eval_result.mp.extent_eou,
                        eval_result.mp.segment_eou,
                    ],
                    hole=0.5,
                    textposition="outside",
                    textinfo="percent+label",
                    marker=dict(colors=["#8ACAA1", "#FFE4B5", "#F7ADAA", "#dd3f3f"]),
                ),
                row=row,
                col=col,
            )
            
            text = f"[{idx}] {eval_result.name[:7]}"
            text += "..." if len(eval_result.name) > 7 else ""
            annotations.append(
                dict(
                    text=text,
                    x=sum(fig.get_subplot(row, col).x) / 2,
                    y=sum(fig.get_subplot(row, col).y) / 2,
                    showarrow=False,
                    xanchor="center",
                )
            )
        fig.update_layout(annotations=annotations)

        return fig



# coding: utf-8

import os
import queue
import time
from collections import namedtuple
from copy import deepcopy
from threading import Thread

import multiprocessing as mp

from supervisely import logger
from supervisely.annotation.annotation import Annotation
from supervisely.imaging import image as sly_image
from supervisely.nn.legacy.config import AlwaysPassingConfigValidator
from supervisely.project.project import Project, read_single_project, OpenMode
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.paths import TaskPaths
from supervisely.task.progress import report_inference_finished
from supervisely.nn.legacy.hosted.inference_modes import InferenceModeFactory
from supervisely.nn.legacy.hosted.inference_batch import determine_task_inference_mode_config
from supervisely.task.progress import Progress


InferenceRequest = namedtuple('InferenceRequest', ['ds_name', 'item_name', 'item_paths'])
InferenceResponse = namedtuple('InferenceResponse', ['ds_name', 'item_name', 'item_paths', 'ann_json', 'meta_json'])


def single_inference_process_fn(inference_initializer, inference_mode_config, in_project_meta_json, request_queue,
                                result_meta_queue, progress_queue, project):
    """Loads a separate model, processes requests from request_queue, results go to result_queue.

    None request signals the process to finish.
    """
    single_image_inference = inference_initializer()
    inference_mode = InferenceModeFactory.create(
        inference_mode_config, ProjectMeta.from_json(in_project_meta_json), single_image_inference)

    project_meta_sent = False
    req = ''
    while req is not None:
        req = request_queue.get()
        if req is not None:
            # Send the resulting project meta to the parent project to make sure we only write the meta JSON once.
            if not project_meta_sent:
                try:
                    result_meta_queue.put(inference_mode.out_meta.to_json(), block=False)
                except queue.Full:
                    pass
            project_meta_sent = True

            in_ann = Annotation.load_json_file(req.item_paths.ann_path, inference_mode.out_meta)
            ann = inference_mode.infer_annotate_image_file(req.item_paths.img_path, in_ann)
            out_dataset = project.datasets.get(req.ds_name)
            out_dataset.add_item_file(
                req.item_name, req.item_paths.img_path, ann=ann, _validate_item=False, _use_hardlink=True)
            progress_queue.put(1)


def progress_report_thread_fn(in_project, progress_queue):
    """Gets inference result annotations from the queue and writes them to the output dataset.
    None result signals the thread to finish.
    """

    progress_bar = Progress('Model applying: ', in_project.total_items)
    while True:
        resp = progress_queue.get()
        if resp is not None:
            progress_bar.iter_done_report()
        else:
            break


def populate_inference_requests_queue(in_project, inference_processes, request_queue):
    for in_dataset in in_project:
        for in_item_name in in_dataset:
            logger.trace('Will process image',
                         extra={'dataset_name': in_dataset.name, 'image_name': in_item_name})
            in_item_paths = in_dataset.get_item_paths(in_item_name)
            req = InferenceRequest(ds_name=in_dataset.name, item_name=in_item_name,
                                   item_paths=in_item_paths)
            while True:
                try:
                    # Set a finite timeout for the queue insertion attempt to prevent deadlocks if all of the
                    # inference processes die in the interim.
                    request_queue.put(req, timeout=0.1)
                    break
                except queue.Full:
                    # If any of the inference processes has died, stop populating the requests queue and exit right
                    # away. Otherwise a deadlock is possible if no inference processes survive to take requests off
                    # the queue.
                    if not all(p.is_alive() for p in inference_processes):
                        # Early exit, return False to indicate failure.
                        return False
    return True


class BatchInferenceMultiprocessApplier:
    """Runs batch inference using multiple worker processes; distributes requests via queue."""
    QUEUE_ELEMENTS_PER_PROCESS = 40

    def __init__(self, single_image_inference_initializer, num_processes, default_inference_mode_config: dict,
                 config_validator=None):
        """
        :param single_image_inference_initializer: Callable returning inference instance.
        :type single_image_inference_initializer: Callable
        :param num_processes: Worker count.
        :type num_processes: int
        :param default_inference_mode_config: Default mode config.
        :type default_inference_mode_config: dict
        :param config_validator: Optional validator.
        :type config_validator: Optional[ConfigValidator]
        """
        self._config_validator = config_validator or AlwaysPassingConfigValidator()
        self._inference_mode_config = determine_task_inference_mode_config(deepcopy(default_inference_mode_config))

        self._in_project = read_single_project(TaskPaths.DATA_DIR)
        logger.info('Project structure has been read. Samples: {}.'.format(self._in_project.total_items))

        out_dir = os.path.join(TaskPaths.RESULTS_DIR, self._in_project.name)
        self._out_project = Project(out_dir, OpenMode.CREATE)
        # Create the output datasets in advance so that the worker processes do not collide on creating their own on
        # disk.
        for in_dataset in self._in_project:
            self._out_project.create_dataset(in_dataset.name)

        self._inference_request_queue = mp.Queue(maxsize=self.QUEUE_ELEMENTS_PER_PROCESS * num_processes)
        self._result_meta_queue = mp.Queue(maxsize=3)
        self._progress_report_queue = mp.Queue(maxsize=self.QUEUE_ELEMENTS_PER_PROCESS * num_processes)
        self._inference_processes = [
            mp.Process(
                target=single_inference_process_fn,
                args=(single_image_inference_initializer, self._inference_mode_config, self._in_project.meta.to_json(),
                      self._inference_request_queue, self._result_meta_queue,
                      self._progress_report_queue, self._out_project),
                daemon=True)
            for _ in range(num_processes)]
        logger.info('Dataset inference preparation done.')
        for p in self._inference_processes:
            p.start()

    def run_inference(self):
        progress_report_thread = Thread(
            target=progress_report_thread_fn, args=(self._in_project, self._progress_report_queue), daemon=True)
        progress_report_thread.start()

        feed_status = populate_inference_requests_queue(
            self._in_project, self._inference_processes, self._inference_request_queue)
        for _ in self._inference_processes:
            self._inference_request_queue.put(None)

        out_meta_json = self._result_meta_queue.get()
        self._out_project.set_meta(ProjectMeta.from_json(out_meta_json))

        for p in self._inference_processes:
            p.join()

        if not feed_status or not all(p.exitcode == 0 for p in self._inference_processes):
            raise RuntimeError('One of the inference processes encountered an error.')

        self._progress_report_queue.put(None)
        progress_report_thread.join()
        report_inference_finished()



from io import StringIO
from pathlib import Path
from typing import List, Optional, Dict

from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget, Editor

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def initialize():
    try:
        # imports here because ordinar import cause error https://github.com/supervisely/issues/issues/1872
        from ruamel.yaml import YAML
        from ruamel.yaml.comments import CommentedMap
    except ModuleNotFoundError:
        raise ModuleNotFoundError('This dependency not provided by Supervisely SDK.\nPlease, install it manually if nedeed.\npip install ruamel.yaml')
    
    class MyYAML(YAML):
        """Small wrapper around ruamel YAML that can return a dumped string when no stream is provided."""

        def dump(self, data, stream=None, **kw):
            inefficient = False
            if stream is None:
                inefficient = True
                stream = StringIO()
            YAML.dump(self, data, stream, **kw)
            if inefficient:
                return stream.getvalue()
    return MyYAML(), CommentedMap

class TabsDynamic(Widget):
    """Widget that renders YAML fragments in editable tabs and can merge them back into a single YAML document."""

    class TabPane:
        """One tab pane (label + content widget)."""

        def __init__(self, label: str, content: Widget):
            """:param label: Tab label.
            :type label: str
            :param content: Widget content for this tab.
            :type content: Widget
            """
            self.label = label
            self.name = label  # identifier corresponding to the active tab
            self.content = content

    def __init__(
        self,
        filepath_or_raw_yaml: str,
        type: Optional[Literal["card", "border-card"]] = "border-card",
        disabled: Optional[bool] = False,
        widget_id=None,
    ):
        """:param filepath_or_raw_yaml: Path to YAML file or raw YAML string.
        :type filepath_or_raw_yaml: str
        :param type: Style: "card" or "border-card".
        :type type: Literal["card", "border-card"], optional
        :param disabled: If True, editors are read-only.
        :type disabled: bool, optional
        :param widget_id: Unique widget identifier.

        :raises ValueError: If YAML is invalid or not dict-like.
        """
        self._disabled = disabled
        try:
            with open(filepath_or_raw_yaml, "r") as file:
                data_source = file.read()
        except:
            data_source = filepath_or_raw_yaml

        yaml, CommentedMap = initialize()
        self._data = yaml.load(data_source)
        if not isinstance(self._data, (dict, CommentedMap)):
            raise ValueError("Invalid yaml path or file content.")
        self._common_data = self._data.copy()
        
        self._items_dict = {}
        self._items = []
        for label, yaml_fragment in self._data.items():
            if isinstance(yaml_fragment, CommentedMap):
                yaml_str = yaml.dump(yaml_fragment)
                editor = Editor(yaml_str, language_mode='yaml', height_px=250)
                if self._disabled:
                    editor.readonly = True
                self._items_dict[label] = editor
                self._items.append(TabsDynamic.TabPane(label=label, content=editor))
                del self._common_data[label]


        if len(self._common_data) > 0:
            yaml_str = yaml.dump(self._common_data)
            editor = Editor(yaml_str, language_mode='yaml', height_px=250)
            if self._disabled:
                editor.readonly = True
            self._items_dict['hyparameters'] = editor
            self._items.append(TabsDynamic.TabPane(label='hyparameters', content=editor))

        assert len(set(self._items_dict.keys())) == len(self._items_dict.keys()), ValueError("All of tab labels should be unique.")
        assert len(self._items_dict.keys()) == len(self._items), ValueError("labels length must be equal to contents length in Tabs widget.")

        self._value = list(self._items_dict.keys())[0]
        self._type = type
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "type": self._type,
            "disabled": self._disabled
        }

    def get_json_state(self) -> Dict:
        return {"value": self._value}

    def set_active_tab(self, value: str):
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_active_tab(self) -> str:
        return StateJson()[self.widget_id]["value"]
    
    def get_merged_yaml(self, as_dict: bool = False):
        yaml, CommentedMap = initialize()
        yaml_data = yaml.load('hyparameters:')
        for label, editor in self._items_dict.items():
            label_yaml_data = yaml.load(editor.get_text())
            if label == 'hyparameters':
                for key, value in label_yaml_data.items():
                    yaml_data[key] = value
            else:
                yaml_data[label] = label_yaml_data
        del yaml_data['hyparameters']
        if as_dict:
            return yaml_data
        return yaml.dump(yaml_data)
    
    def disable(self):
        self._disabled = True
        for key, editor in self._items_dict.items():
            editor.readonly = True
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        for key, editor in self._items_dict.items():
            editor.readonly = False
        self._disabled = False
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()


from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api
from supervisely.app.widgets_context import JinjaWidgets


class VideoPlayer(Widget):
    """Video player widget with play/pause and seek; displays video from URL with configurable MIME type."""

    def __init__(self, url: str = None, mime_type: str = "video/mp4", widget_id: str = None):
        """:param url: Video URL.
        :type url: str, optional
        :param mime_type: MIME type (e.g. "video/mp4").
        :type mime_type: str
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._api = Api()

        self._url = url
        self._mime_type = mime_type
        self._mask_path = None

        self._current_timestamp = 0
        self._is_playing = False

        super().__init__(widget_id=widget_id, file_path=__file__)

        script_path = "./sly/css/app/widgets/video_player/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def get_json_data(self):
        return {
            "url": self._url,
            "mimeType": self._mime_type,
            "maskPath": self._mask_path,
        }

    def get_json_state(self):
        return {
            "currentTime": 0,
            "timeToSet": 0,
            "isPlaying": False,
        }

    @property
    def url(self):
        return self._url

    @property
    def mime_type(self):
        return self._mime_type

    def set_video(self, url: str, mime_type: str = "video/mp4"):
        self._url = url
        self._mime_type = mime_type
        DataJson()[self.widget_id]["url"] = self._url
        DataJson()[self.widget_id]["mimeType"] = self._mime_type
        DataJson().send_changes()
        StateJson()[self.widget_id]["currentTime"] = 0
        StateJson().send_changes()

    def play(self):
        is_playing = StateJson()[self.widget_id]["isPlaying"]
        if is_playing is True:
            return
        self._is_playing = True
        StateJson()[self.widget_id]["isPlaying"] = True
        StateJson().send_changes()

    def pause(self):
        is_playing = StateJson()[self.widget_id]["isPlaying"]
        if is_playing is False:
            return
        self._is_playing = False
        StateJson()[self.widget_id]["isPlaying"] = False
        StateJson().send_changes()

    def get_current_timestamp(self):
        self._current_timestamp = round(StateJson()[self.widget_id]["currentTime"], 1)
        return self._current_timestamp

    def set_current_timestamp(self, value: int):
        self._current_timestamp = value
        StateJson()[self.widget_id]["timeToSet"] = value
        StateJson().send_changes()
        return self._current_timestamp

    def draw_mask(self, path):
        self._mask_path = path
        DataJson()[self.widget_id]["maskPath"] = self._mask_path
        DataJson().send_changes()

    def hide_mask(self):
        self._mask_path = None
        DataJson()[self.widget_id]["maskPath"] = self._mask_path
        DataJson().send_changes()



from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class RTDETR(BaseTrainArtifacts):
    """Train/serve artifacts handler for RT-DETR object detection models."""

    def __init__(self, team_id: int):
        """See :class:`~supervisely.nn.artifacts.artifacts.BaseTrainArtifacts` for params."""
        super().__init__(team_id)

        self._app_name = "Train RT-DETR"
        self._slug = "supervisely-ecosystem/rt-detr/supervisely_integration/train"
        self._serve_app_name = "Serve RT-DETR"
        self._serve_slug = "supervisely-ecosystem/rt-detr/supervisely_integration/serve"
        self._framework_name = "RT-DETR"
        self._framework_folder = "/RT-DETR"
        self._weights_folder = "weights"
        self._task_type = "object detection"
        self._weights_ext = ".pth"
        self._config_file = "config.yml"
        self._pattern = re_compile(r"^/RT-DETR/[^/]+/\d+/?$")
        self._available_task_types: List[str] = ["object detection"]
        self._require_runtime = False
        self._has_benchmark_evaluation = True

    def get_task_id(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[-1]

    def get_project_name(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[-2]

    def get_task_type(self, artifacts_folder: str) -> str:
        return self._task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._config_file)



from supervisely.project.versioning.video_schema import VideoSnapshotSchema, _VIDEO_SCHEMAS
from supervisely.project.versioning.volume_schema import VolumeSnapshotSchema, _VOLUME_SCHEMAS

DEFAULT_IMAGE_SCHEMA_VERSION = "v1.0.0"
DEFAULT_VOLUME_SCHEMA_VERSION = "v2.0.0"
DEFAULT_VIDEO_SCHEMA_VERSION = "v2.0.0"


def get_video_snapshot_schema(schema_version: str) -> VideoSnapshotSchema:
    schema = _VIDEO_SCHEMAS.get(schema_version)
    if schema is None:
        raise RuntimeError(f"Unsupported video snapshot schema_version: {schema_version!r}")
    return schema


def get_volume_snapshot_schema(schema_version: str) -> VolumeSnapshotSchema:
    schema = _VOLUME_SCHEMAS.get(schema_version)
    if schema is None:
        raise RuntimeError(f"Unsupported volume snapshot schema_version: {schema_version!r}")
    return schema



import copy
import time
import uuid
from collections import defaultdict
from typing import List, Union

from supervisely.project.project_meta import ProjectMeta
from supervisely.annotation.annotation import Annotation
from supervisely.api.annotation_api import AnnotationInfo
from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.project.project_meta import ProjectMeta


class GridGalleryV2(Widget):
    """Grid gallery widget (v2) that renders images with annotations, filters, and object bindings."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        IMAGE_CLICKED = "image_clicked_cb"

    def __init__(
        self,
        columns_number: int,
        fit_on_resize: bool = True,
        enable_zoom: bool = True,
        show_opacity_slider: bool = True,
        show_zoom_slider: bool = True,
        annotations_opacity: float = 0.8,
        enable_pointer_events: bool = True,
        transparent_background: bool = True,
        show_filter: bool = True,
        fill_rectangle: bool = True,
        enable_panning=False,
        border_width: int = 3,
        default_tag_filters: List[Union[str, dict]] = None,
        widget_id: str = None,
    ):
        """
        :param columns_number: Number of columns in the grid.
        :type columns_number: int
        :param fit_on_resize: If True, fit images on resize.
        :type fit_on_resize: bool
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool
        :param show_opacity_slider: If True, show opacity slider.
        :type show_opacity_slider: bool
        :param show_zoom_slider: If True, show zoom slider.
        :type show_zoom_slider: bool
        :param annotations_opacity: Opacity of annotation overlays (0–1).
        :type annotations_opacity: float
        :param enable_pointer_events: If True, enable pointer events on overlays.
        :type enable_pointer_events: bool
        :param transparent_background: If True, use transparent background.
        :type transparent_background: bool
        :param show_filter: If True, show tag filter.
        :type show_filter: bool
        :param fill_rectangle: If True, fill rectangles.
        :type fill_rectangle: bool
        :param enable_panning: If True, enable panning.
        :type enable_panning: bool
        :param border_width: Border width for shapes.
        :type border_width: int
        :param default_tag_filters: Default tag filters applied.
        :type default_tag_filters: List[Union[str, dict]], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._data = []
        self._layout = []
        self._annotations = {}

        self.columns_number = columns_number

        self._last_used_column_index = 0
        self._project_meta: ProjectMeta = None
        self._loading = False

        self._bindings_dict = defaultdict(list)

        #############################
        # grid gallery settings
        self._show_preview: bool = True
        self._fill_rectangle: bool = fill_rectangle
        self._border_width: int = border_width

        self._fit_on_resize: bool = fit_on_resize
        self._enable_zoom: bool = enable_zoom
        self._show_opacity_header: bool = show_opacity_slider
        self._show_zoom_header: bool = show_zoom_slider
        self._opacity: float = annotations_opacity
        self._enable_pointer_events: bool = enable_pointer_events
        self._transparent_background: bool = transparent_background
        self._show_filter: bool = show_filter

        self._enablePan = enable_panning

        self._filters_tags = default_tag_filters
        #############################

        self._filters = []
        self._object_bindings = []

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _generate_project_meta(self):
        objects_dict = dict()
        obj_tags_dict = dict()

        for cell_data in self._data:
            ann_info: AnnotationInfo = cell_data["annotation_info"]
            project_meta: ProjectMeta = cell_data["project_meta"]

            # tmp = project_meta
            # if self._project_meta is not None:  # TODO
            #     project_meta = self._project_meta

            annotation = Annotation.from_json(ann_info.annotation, project_meta)
            for idx, label in enumerate(annotation.labels):
                objects_dict[label.obj_class.name] = label.obj_class
                for tag in label.tags:
                    obj_tags_dict[tag.name] = project_meta.get_tag_meta(tag.name)

        objects_list = list(objects_dict.values())
        tags_list = list(obj_tags_dict.values())

        self._project_meta = ProjectMeta(obj_classes=objects_list, tag_metas=tags_list)
        return self._project_meta.to_json()

    def get_json_data(self):
        return {
            "content": {
                "projectMeta": self._generate_project_meta(),
                "layout": self._layout,
                "annotations": self._annotations,
            },
            "loading": self._loading,
        }

    def get_json_state(self):
        return {
            "filters": self._filters,
            "objectBindings": self._object_bindings,
            "options": {
                "fitOnResize": self._fit_on_resize,
                "enableZoom": self._enable_zoom,
                "showOpacityInHeader": self._show_opacity_header,
                "showZoomInHeader": self._show_zoom_header,
                "opacity": self._opacity,
                "enableObjectsPointerEvents": self._enable_pointer_events,
                "showTransparentBackground": self._transparent_background,
                "showFilter": self._show_filter,
                "enablePan": self._enablePan,
                "lineWidth": self._border_width,
            },
        }

    def get_column_index(self, incoming_value):
        if incoming_value is not None and 0 > incoming_value > self.columns_number:
            raise ValueError(f"column index == {incoming_value} is out of bounds")

        if incoming_value is None:
            incoming_value = self._last_used_column_index
            self._last_used_column_index = (self._last_used_column_index + 1) % self.columns_number
        else:
            self._last_used_column_index = incoming_value

        return incoming_value

    def append(
        self,
        image_url: str,
        annotation_info: AnnotationInfo,
        project_meta: ProjectMeta,
        title: str = "",
        column_index: int = None,
        ignore_tags_filtering: Union[bool, List[str]] = False,
        call_update: bool = True,
    ):
        column_index = self.get_column_index(incoming_value=column_index)
        cell_uuid = str(
            uuid.uuid5(
                namespace=uuid.NAMESPACE_URL,
                name=f"{image_url}_{title}_{column_index}_{time.time()}",
            ).hex
        )

        if ignore_tags_filtering is True:
            pass

        self._data.append(
            {
                "image_url": image_url,
                "annotation_info": annotation_info,
                "project_meta": project_meta,
                "column_index": column_index,
                "cell_uuid": cell_uuid,
                "skipObjectTagsFiltering": ignore_tags_filtering,

            }
        )

        if call_update:
            self._update()
        return cell_uuid

    def clean_up(self):
        self._data = []
        self._layout = []
        self._annotations = {}
        self._update()
        self.update_data()

    def _update_layout(self):
        layout = [[] for _ in range(self.columns_number)]

        for cell_data in self._data:
            tmp = cell_data["cell_uuid"]
            skip_filters: Union[bool, List[str]] = cell_data.get("skipObjectTagsFiltering")
            if skip_filters is True or isinstance(skip_filters, list):
                tmp = {
                    "layoutDataKey": cell_data["cell_uuid"],
                    "options": {"skipObjectTagsFiltering": skip_filters},
                }

            layout[cell_data["column_index"]].append(tmp)

        self._layout = copy.deepcopy(layout)
        DataJson()[self.widget_id]["content"]["layout"] = self._layout

    def _update_annotations(self):
        annotations = {}
        for cell_data in self._data:
            annotations[cell_data["cell_uuid"]] = {
                "imageUrl": cell_data["image_url"],
                "annotation": cell_data["annotation_info"]._asdict(),
            }
        self._annotations = copy.deepcopy(annotations)
        DataJson()[self.widget_id]["content"]["annotations"] = self._annotations

    def _update_object_bindings(self):
        object_bindings = []
        for cell_data in self._data:
            ann_json = cell_data["annotation_info"].annotation
            for obj in ann_json["objects"]:
                self._bindings_dict[obj["classTitle"]].append(
                    {"id": obj["id"], "annotationKey": cell_data["cell_uuid"]}
                )

        def filter_unique_ids(dict_list):
            seen_ids = set()
            seen_anns = set()
            unique_dicts = []

            for item in dict_list:
                if item["id"] not in seen_ids:
                    seen_ids.add(item["id"])
                    unique_dicts.append(item)

            return unique_dicts

        for class_name, bindings in self._bindings_dict.items():
            object_bindings.append(filter_unique_ids(bindings))

        self._object_bindings = object_bindings
        DataJson()[self.widget_id]["content"]["objectsBindings"] = object_bindings

    def _update_project_meta(self):
        DataJson()[self.widget_id]["content"]["projectMeta"] = self._generate_project_meta()

    def _update_filters(self):
        filters = []
        if self._filters_tags is not None:
            for filters_tag in self._filters_tags:
                tmp = list(filters_tag.items())[0]
                filters.append({"type": "tag", "tagId": tmp[0], "value": tmp[1]})
        self._filters = filters
        StateJson()[self.widget_id]["filters"] = filters

    def _update(self):
        self._update_layout()
        self._update_annotations()
        self._update_object_bindings()
        self._update_project_meta()
        self._update_filters()

        DataJson().send_changes()
        StateJson().send_changes()

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()



import numpy as np

pycocotools_installed = False
try:
    from pycocotools.coco import COCO  # pylint: disable=import-error
    from pycocotools.cocoeval import COCOeval  # pylint: disable=import-error

    pycocotools_installed = True
except ImportError:
    COCO = object
    COCOeval = object


class SlyCOCOeval(COCOeval):
    """COCO evaluation wrapper with customized summarize output for Supervisely benchmark reports."""

    def summarize(self):
        """
        Compute and display summary metrics for evaluation results.
        Note this functin can *only* be applied on the default parameter setting
        """

        def _summarize(ap=1, iouThr=None, areaRng="all", maxDets=100):
            p = self.params
            iStr = " {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}"
            titleStr = "Average Precision" if ap == 1 else "Average Recall"
            typeStr = "(AP)" if ap == 1 else "(AR)"
            iouStr = (
                "{:0.2f}:{:0.2f}".format(p.iouThrs[0], p.iouThrs[-1])
                if iouThr is None
                else "{:0.2f}".format(iouThr)
            )

            aind = [i for i, aRng in enumerate(p.areaRngLbl) if aRng == areaRng]
            mind = [i for i, mDet in enumerate(p.maxDets) if mDet == maxDets]
            if ap == 1:
                # dimension of precision: [TxRxKxAxM]
                s = self.eval["precision"]
                # IoU
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]
                s = s[:, :, :, aind, mind]
            else:
                # dimension of recall: [TxKxAxM]
                s = self.eval["recall"]
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]
                s = s[:, :, aind, mind]
            if len(s[s > -1]) == 0:
                mean_s = -1
            else:
                mean_s = np.mean(s[s > -1])
            print(iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s))
            return mean_s

        def _summarizeDets():
            stats = np.zeros((12,))
            stats[0] = _summarize(1, maxDets=self.params.maxDets[2])
            stats[1] = _summarize(1, iouThr=0.5, maxDets=self.params.maxDets[2])
            stats[2] = _summarize(1, iouThr=0.75, maxDets=self.params.maxDets[2])
            stats[3] = _summarize(1, areaRng="small", maxDets=self.params.maxDets[2])
            stats[4] = _summarize(1, areaRng="medium", maxDets=self.params.maxDets[2])
            stats[5] = _summarize(1, areaRng="large", maxDets=self.params.maxDets[2])
            stats[6] = _summarize(0, maxDets=self.params.maxDets[0])
            stats[7] = _summarize(0, maxDets=self.params.maxDets[1])
            stats[8] = _summarize(0, maxDets=self.params.maxDets[2])
            stats[9] = _summarize(0, areaRng="small", maxDets=self.params.maxDets[2])
            stats[10] = _summarize(0, areaRng="medium", maxDets=self.params.maxDets[2])
            stats[11] = _summarize(0, areaRng="large", maxDets=self.params.maxDets[2])
            return stats

        def _summarizeKps():
            stats = np.zeros((10,))
            stats[0] = _summarize(1, maxDets=20)
            stats[1] = _summarize(1, maxDets=20, iouThr=0.5)
            stats[2] = _summarize(1, maxDets=20, iouThr=0.75)
            stats[3] = _summarize(1, maxDets=20, areaRng="medium")
            stats[4] = _summarize(1, maxDets=20, areaRng="large")
            stats[5] = _summarize(0, maxDets=20)
            stats[6] = _summarize(0, maxDets=20, iouThr=0.5)
            stats[7] = _summarize(0, maxDets=20, iouThr=0.75)
            stats[8] = _summarize(0, maxDets=20, areaRng="medium")
            stats[9] = _summarize(0, maxDets=20, areaRng="large")
            return stats

        if not self.eval:
            raise Exception("Please run accumulate() first")
        iouType = self.params.iouType
        if iouType == "segm" or iouType == "bbox":
            summarize = _summarizeDets
        elif iouType == "keypoints":
            summarize = _summarizeKps
        self.stats = summarize()  # pylint: disable=possibly-used-before-assignment



# coding: utf-8
"""Create and manage Guides for labeling jobs and labeling queues."""

# docs
from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase


class GuideInfo(NamedTuple):
    """
    Information about a Guide.

    :param id: Guide ID in Supervisely.
    :type id: int
    :param name: Guide name.
    :type name: str
    :param description: Guide description.
    :type description: str
    :param file_path: Path to the guide file (PDF or other).
    :type file_path: str
    :param created_at: Guide creation date.
    :type created_at: str
    :param updated_at: Guide last update date.
    :type updated_at: str
    :param created_by_id: ID of the User who created the Guide.
    :type created_by_id: int
    :param team_id: Team ID where the Guide is located.
    :type team_id: int
    :param video_id: ID of the video associated with the guide (if any).
    :type video_id: Optional[int]
    :param disabled_by: ID of the User who disabled the Guide (if disabled).
    :type disabled_by: Optional[int]
    :param disabled_at: Date when the Guide was disabled (if disabled).
    :type disabled_at: Optional[str]
    """

    id: int
    name: str
    description: str
    file_path: str
    created_at: str
    updated_at: str
    created_by_id: int
    team_id: int
    video_id: Optional[int] = None
    disabled_by: Optional[int] = None
    disabled_at: Optional[str] = None


class GuidesApi(ModuleApiBase):
    """API for working with Guides."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                guides = api.guides.get_list(team_id=123)
        """
        super().__init__(api)

    @staticmethod
    def info_sequence():
        """
        Sequence of fields that are returned by the API to represent GuideInfo.

        :Usage Example:

            .. code-block:: python

                GuideInfo(
                    id=1,
                    name="How to label objects",
                    description="Comprehensive guide on object labeling",
                    file_path="/path/to/guide.pdf",
                    created_at="2023-01-01T00:00:00.000Z",
                    updated_at="2025-11-17T18:21:10.217Z",
                    created_by_id=1,
                    team_id=1,
                    video_id=None,
                    disabled_by=None,
                    disabled_at=None,
                )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.FILE_PATH,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.CREATED_BY_ID,
            ApiField.TEAM_ID,
            ApiField.VIDEO_ID,
            ApiField.DISABLED_BY,
            ApiField.DISABLED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Name of the tuple that represents GuideInfo.
        """
        return "GuideInfo"

    def get_list(
        self, team_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[GuideInfo]:
        """
        List of Guides in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of parameters to filter Guides.
        :type filters: List[Dict[str, str]], optional
        :returns: List of information about Guides.
        :rtype: List[:class:`~supervisely.api.guides_api.GuideInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_id = 123
                guides = api.guides.get_list(team_id)
                print(guides)
                # Output: [
                #     GuideInfo(
                #         id=1,
                #         name='How to label objects',
                #         description='Comprehensive guide on object labeling',
                #         file_path='/path/to/guide.pdf',
                #         created_at='2023-01-01T00:00:00.000Z',
                #         updated_at='2025-11-17T18:21:10.217Z',
                #         created_by_id=1,
                #         team_id=1,
                #         video_id=None,
                #         disabled_by=None,
                #         disabled_at=None
                #     )
                # ]
        """
        return self.get_list_all_pages(
            "guides.list",
            {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []},
        )

    def get_info_by_id(self, id: int) -> GuideInfo:
        """
        Get Guide information by ID.

        :param id: Guide ID in Supervisely.
        :type id: int
        :returns: Information about Guide.
        :rtype: :class:`~supervisely.api.guides_api.GuideInfo`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                guide_id = 1
                guide_info = api.guides.get_info_by_id(guide_id)
                print(guide_info)
                # Output: GuideInfo(
                #     id=1,
                #     name='How to label objects',
                #     description='Comprehensive guide on object labeling',
                #     file_path='/path/to/guide.pdf',
                #     created_at='2023-01-01T00:00:00.000Z',
                #     updated_at='2025-11-17T18:21:10.217Z',
                #     created_by_id=1,
                #     team_id=1,
                #     video_id=None,
                #     disabled_by=None,
                #     disabled_at=None
                # )
        """
        return self._get_info_by_id(id, "guides.info")
